# yapf: disable
# isort: skip_file
# ruff: noqa

# To not generate code where imported names might get shadowed when a user
# specifies some name in their proto file to be the same as one of our imported
# names, (for example: a request field named `uuid`) we bind all imports to
# names that are forbidden in 'proto' and therefore can never collide.

# Standard imports.
from __future__ import annotations as IMPORT_future_annotations

# The following MUST appear before the rest of the imports, since those imports
# may be invalid (broken) if the generated code is mismatched with the installed
# libraries.
import reboot.versioning as IMPORT_reboot_versioning
IMPORT_reboot_versioning.check_generated_code_compatible("0.25.5")

# ATTENTION: no types in this file should be imported with their unqualified
#            name (e.g. `from typing import Any`). That would cause clashes
#            with user-defined methods that have the same name. Use
#            fully-qualified names (e.g. `IMPORT_typing.Any`) instead.
import asyncio as IMPORT_asyncio
import dataclasses as IMPORT_dataclasses
import google.protobuf.descriptor as IMPORT_google_protobuf_descriptor
import google.protobuf.json_format as IMPORT_google_protobuf_json_format
import google.protobuf.message as IMPORT_google_protobuf_message
import grpc as IMPORT_grpc
import grpc_status._async as IMPORT_rpc_status_async
from grpc_status import rpc_status as IMPORT_rpc_status_sync
import json as IMPORT_json
import os as IMPORT_os
import traceback as IMPORT_traceback
import uuid as IMPORT_uuid
import pickle as IMPORT_pickle
import reboot as IMPORT_reboot
import log.log as IMPORT_log_log   # type: ignore[import]
import typing as IMPORT_typing
import reboot.aio.backoff as IMPORT_reboot_aio_backoff
import functools as IMPORT_functools
from abc import abstractmethod as IMPORT_abc_abstractmethod
from datetime import datetime as IMPORT_datetime_datetime
from datetime import timedelta as IMPORT_datetime_timedelta
from datetime import timezone as IMPORT_datetime_timezone
from google.protobuf import timestamp_pb2 as IMPORT_google_protobuf_timestamp_pb2
from google.protobuf import wrappers_pb2 as IMPORT_google_protobuf_wrappers_pb2
import reboot.aio.tracing as IMPORT_reboot_aio_tracing
from google.rpc import status_pb2 as IMPORT_google_rpc_status_pb2
from tzlocal import get_localzone as IMPORT_tzlocal_get_localzone
import reboot.aio.call as IMPORT_reboot_aio_call
import reboot.aio.contexts as IMPORT_reboot_aio_contexts
import reboot.aio.headers as IMPORT_reboot_aio_headers
import reboot.aio.idempotency as IMPORT_reboot_aio_idempotency
import reboot.aio.internals.channel_manager as IMPORT_reboot_aio_internals_channel_manager
import reboot.aio.internals.middleware as IMPORT_reboot_aio_internals_middleware
import reboot.aio.internals.tasks_cache as IMPORT_reboot_aio_internals_tasks_cache
import reboot.aio.internals.tasks_dispatcher as IMPORT_reboot_aio_internals_tasks_dispatcher
import reboot.aio.placement as IMPORT_reboot_aio_placement
import reboot.aio.servicers as IMPORT_reboot_aio_servicers
import reboot.aio.state_managers as IMPORT_reboot_aio_state_managers
import reboot.aio.stubs as IMPORT_reboot_aio_stubs
import reboot.aio.tasks as IMPORT_reboot_aio_tasks
import reboot.aio.types as IMPORT_reboot_aio_types
import reboot.aio.external as IMPORT_reboot_aio_external
import reboot.settings as IMPORT_reboot_settings
import reboot.nodejs.python as IMPORT_reboot_nodejs_python
from reboot.time import DateTimeWithTimeZone as IMPORT_reboot_time_DateTimeWithTimeZone
import rbt.v1alpha1 as IMPORT_rbt_v1alpha1
import google.protobuf.any_pb2 as IMPORT_google_protobuf_any_pb2

# User defined or referenced imports.
import rbt.std.collections.v1.sorted_map_pb2
import rbt.std.collections.v1.sorted_map_pb2_grpc
import rbt.v1alpha1.options_pb2
# Additionally re-export all messages and enums from the pb2 module.
from rbt.std.collections.v1.sorted_map_pb2 import (
    Entry,
    GetRequest,
    GetResponse,
    InsertRequest,
    InsertResponse,
    InvalidRangeError,
    RangeRequest,
    RangeResponse,
    RemoveRequest,
    RemoveResponse,
    ReverseRangeRequest,
    ReverseRangeResponse,
)

logger = IMPORT_log_log.get_logger(__name__)


############################ Legacy gRPC Servicers ############################
# This section is relevant (only) for servicers that implement a legacy gRPC
# service in a Reboot context. It is irrelevant to clients.

def MakeLegacyGrpcServiceable(
    # A legacy gRPC servicer type can't be more specific than `type`,
    # because legacy gRPC servicers (as generated by the gRPC `protoc`
    # plugin) do not share any common base class other than `object`.
    servicer_type: type
) -> IMPORT_reboot_aio_servicers.Serviceable:
    raise ValueError(f"Unknown legacy gRPC servicer type '{servicer_type}'")



############################ Reboot Servicer Middlewares ############################
# This section is relevant (only) for servicers implementing a Reboot servicer. It
# is irrelevant to clients, except for the fact that some clients are _also_ such
# servicers.

# For internal calls, we can use a magic token to bypass token verification and
# authorization checks. The token provides no auth information (e.g.,
# `context.auth is None`).
__internal_magic_token__: str = f'internal-{str(IMPORT_uuid.uuid4())}'

class SortedMapServicerMiddleware(IMPORT_reboot_aio_internals_middleware.Middleware):

    def __init__(
        self,
        *,
        servicer: SortedMapServicer,
        application_id: IMPORT_reboot_aio_types.ApplicationId,
        consensus_id: IMPORT_reboot_aio_types.ConsensusId,
        state_manager: IMPORT_reboot_aio_state_managers.StateManager,
        placement_client: IMPORT_reboot_aio_placement.PlacementClient,
        channel_manager: IMPORT_reboot_aio_internals_channel_manager._ChannelManager,
        tasks_cache: IMPORT_reboot_aio_internals_tasks_cache.TasksCache,
        token_verifier: IMPORT_typing.Optional[IMPORT_reboot.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: IMPORT_reboot_aio_contexts.EffectValidation,
        app_internal_api_key_secret: str,
        ready: IMPORT_asyncio.Event,
    ):
        super().__init__(
            application_id=application_id,
            consensus_id=consensus_id,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            service_names = [
                IMPORT_reboot_aio_types.ServiceName("rbt.std.collections.v1.SortedMapMethods"),
            ],
            placement_client=placement_client,
            channel_manager=channel_manager,
            effect_validation=effect_validation,
            app_internal_api_key_secret=app_internal_api_key_secret,
        )

        self._servicer = servicer
        self._state_manager = state_manager
        self.tasks_dispatcher = IMPORT_reboot_aio_internals_tasks_dispatcher.TasksDispatcher(
            application_id=application_id,
            dispatch=self.dispatch,
            tasks_cache=tasks_cache,
            ready=ready,
        )

        # Store the type of each method's request so that stored requests can be
        # deserialized into the correct type.
        self.request_type_by_method_name: dict[str, type[IMPORT_google_protobuf_message.Message]] = {
            'Insert': rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
            'Remove': rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
            'Get': rbt.std.collections.v1.sorted_map_pb2.GetRequest,
            'Range': rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
            'ReverseRange': rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        }

        # Get authorizer, if any, converting from a rule if necessary.
        def convert_authorizer_rule_if_necessary(
            authorizer_or_rule: IMPORT_typing.Optional[
                IMPORT_reboot.aio.auth.authorizers.Authorizer | IMPORT_reboot.aio.auth.authorizers.AuthorizerRule
            ]
        ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer:

            # If no authorizer or rule is provided, return the default
            # authorizer which allows if app internal or allows if in
            # dev mode (and logs some warnings to help the user
            # realize where they are missing authorization).
            if authorizer_or_rule is None:
                return IMPORT_reboot.aio.auth.authorizers.DefaultAuthorizer(
                    'SortedMap'
                )

            if isinstance(authorizer_or_rule, IMPORT_reboot.aio.auth.authorizers.AuthorizerRule):
                return SortedMapAuthorizer(
                    _default=authorizer_or_rule
                )

            return authorizer_or_rule

        self._authorizer = convert_authorizer_rule_if_necessary(
            servicer.authorizer()
        )

        # Create token verifier.
        self._token_verifier: IMPORT_typing.Optional[IMPORT_reboot.aio.auth.token_verifiers.TokenVerifier] = (
            servicer.token_verifier() or token_verifier
        )

        # Since users specify errors as proto messages they can't raise them
        # directly - to do so they have to use the `Aborted` wrapper, which will
        # hold the original proto message. On errors we'll need to check whether
        # such wrappers hold a proto message for a specified error, so we can
        # avoid retrying tasks that complete with a specified error.
        self._specified_errors_by_service_method_name: dict[str, list[str]] = {
            'rbt.std.collections.v1.SortedMapMethods.Range': [
                'rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError',
            ],
            'rbt.std.collections.v1.SortedMapMethods.ReverseRange': [
                'rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError',
            ],
        }


    def add_to_server(self, server: IMPORT_grpc.aio.Server) -> None:
        rbt.std.collections.v1.sorted_map_pb2_grpc.add_SortedMapMethodsServicer_to_server(
            self, server
        )

    async def inspect(self, state_ref: IMPORT_reboot_aio_types.StateRef) -> IMPORT_typing.AsyncIterator[IMPORT_google_protobuf_message.Message]:
        """Implementation of `Middleware.inspect()`."""
        context = self.create_context(
            headers=IMPORT_reboot_aio_headers.Headers(
                application_id=self.application_id,
                state_ref=state_ref,
            ),
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.ReaderContext,
        )

        async with self._state_manager.streaming_reader_idempotency_key(
            context,
            self._servicer.__state_type__,
            authorize=None,
        ) as states:
            async for (state, idempotency_key) in states:
                yield state

    async def react_query(
        self,
        headers: IMPORT_reboot_aio_headers.Headers,
        method: str,
        request_bytes: bytes,
    ) -> IMPORT_typing.AsyncIterator[tuple[IMPORT_typing.Optional[IMPORT_google_protobuf_message.Message], list[IMPORT_uuid.UUID]]]:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes' for each state
        update that creates a different response.

        # The caller (react.py) should have already ensured that this consensus
        # is authoritative for this traffic.
        assert self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        ) == self._consensus_id

        NOTE: only unary reader methods are supported."""
        # Need to define these up here since we can only do that once.
        last_response: IMPORT_typing.Optional[IMPORT_google_protobuf_message.Message] = None
        aggregated_idempotency_keys: list[IMPORT_uuid.UUID] = []
        if method == 'Insert':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                f"Method '{method}' is invalid for servicer SortedMap."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        elif method == 'Remove':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                f"Method '{method}' is invalid for servicer SortedMap."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        elif method == 'Get':

            context = self.create_context(
                headers=headers,
                state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                context_type=IMPORT_reboot_aio_contexts.ReaderContext,
            )

            context.auth = await self._maybe_verify_token(headers)

            request = rbt.std.collections.v1.sorted_map_pb2.GetRequest()
            request.ParseFromString(request_bytes)

            async with self._state_manager.reactively(
                context,
                self._servicer.__state_type__,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Get',
                    headers=headers,
                    auth=context.auth,
                    request=request,
                ),
            ) as states:
                async for (state, idempotency_keys) in states:

                    aggregated_idempotency_keys.extend(idempotency_keys)

                    # Note: This does not do any defensive copying currently:
                    # see https://github.com/reboot-dev/respect/issues/2636.
                    @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
                    async def run__Get(validating_effects: bool) -> IMPORT_google_protobuf_message.Message:
                        return await self.__Get(
                            context,
                            state,
                            request,
                            validating_effects=validating_effects,
                        )

                    response = await run__Get()

                    if last_response != response:
                        yield (response, aggregated_idempotency_keys)
                        last_response = response
                    else:
                        yield (None, aggregated_idempotency_keys)

                    aggregated_idempotency_keys.clear()
        elif method == 'Range':

            context = self.create_context(
                headers=headers,
                state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                context_type=IMPORT_reboot_aio_contexts.ReaderContext,
            )

            context.auth = await self._maybe_verify_token(headers)

            request = rbt.std.collections.v1.sorted_map_pb2.RangeRequest()
            request.ParseFromString(request_bytes)

            async with self._state_manager.reactively(
                context,
                self._servicer.__state_type__,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Range',
                    headers=headers,
                    auth=context.auth,
                    request=request,
                ),
            ) as states:
                async for (state, idempotency_keys) in states:

                    aggregated_idempotency_keys.extend(idempotency_keys)

                    # Note: This does not do any defensive copying currently:
                    # see https://github.com/reboot-dev/respect/issues/2636.
                    @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
                    async def run__Range(validating_effects: bool) -> IMPORT_google_protobuf_message.Message:
                        return await self.__Range(
                            context,
                            state,
                            request,
                            validating_effects=validating_effects,
                        )

                    response = await run__Range()

                    if last_response != response:
                        yield (response, aggregated_idempotency_keys)
                        last_response = response
                    else:
                        yield (None, aggregated_idempotency_keys)

                    aggregated_idempotency_keys.clear()
        elif method == 'ReverseRange':

            context = self.create_context(
                headers=headers,
                state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                context_type=IMPORT_reboot_aio_contexts.ReaderContext,
            )

            context.auth = await self._maybe_verify_token(headers)

            request = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest()
            request.ParseFromString(request_bytes)

            async with self._state_manager.reactively(
                context,
                self._servicer.__state_type__,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.ReverseRange',
                    headers=headers,
                    auth=context.auth,
                    request=request,
                ),
            ) as states:
                async for (state, idempotency_keys) in states:

                    aggregated_idempotency_keys.extend(idempotency_keys)

                    # Note: This does not do any defensive copying currently:
                    # see https://github.com/reboot-dev/respect/issues/2636.
                    @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
                    async def run__ReverseRange(validating_effects: bool) -> IMPORT_google_protobuf_message.Message:
                        return await self.__ReverseRange(
                            context,
                            state,
                            request,
                            validating_effects=validating_effects,
                        )

                    response = await run__ReverseRange()

                    if last_response != response:
                        yield (response, aggregated_idempotency_keys)
                        last_response = response
                    else:
                        yield (None, aggregated_idempotency_keys)

                    aggregated_idempotency_keys.clear()
        else:
            logger.warning(
                "Got a React query request with an invalid method name: "
                f"Method '{method}' is invalid for servicer SortedMap."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )
            yield  # Unreachable but necessary for mypy.

    async def react_mutate(
        self,
        headers: IMPORT_reboot_aio_headers.Headers,
        method: str,
        request_bytes: bytes,
    ) -> IMPORT_google_protobuf_message.Message:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes'."""
        if method == 'Insert':
            request = rbt.std.collections.v1.sorted_map_pb2.InsertRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `IMPORT_grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            call_backoff = IMPORT_reboot_aio_backoff.Backoff()
            while True:
                # We make a full-fledged gRPC call, so that if this traffic
                # was misrouted (i.e. this consensus is not authoritative
                # for the state), it will now go to the right place. The
                # receiving middleware will handle things like effect
                # validation and so forth.
                assert headers.application_id is not None  # Guaranteed by `Headers`.
                stub = rbt.std.collections.v1.sorted_map_pb2_grpc.SortedMapMethodsStub(
                    self.channel_manager.get_channel_to(
                        self.placement_client.address_for_actor(
                            headers.application_id,
                            headers.state_ref,
                        )
                    )
                )
                call = stub.Insert(
                    request=request,
                    metadata=headers.to_grpc_metadata(),
                )
                try:
                    return await call
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.UNAVAILABLE:
                        await call_backoff()
                        continue

                    # Reconstitute the error that the server threw, if it was a declared error.
                    status = await IMPORT_rpc_status_async.from_call(call)
                    if status is not None:
                        raise SortedMap.InsertAborted.from_status(
                            status
                        ) from None
                    raise SortedMap.InsertAborted.from_grpc_aio_rpc_error(
                        error
                     ) from None

        elif method == 'Remove':
            request = rbt.std.collections.v1.sorted_map_pb2.RemoveRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `IMPORT_grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            call_backoff = IMPORT_reboot_aio_backoff.Backoff()
            while True:
                # We make a full-fledged gRPC call, so that if this traffic
                # was misrouted (i.e. this consensus is not authoritative
                # for the state), it will now go to the right place. The
                # receiving middleware will handle things like effect
                # validation and so forth.
                assert headers.application_id is not None  # Guaranteed by `Headers`.
                stub = rbt.std.collections.v1.sorted_map_pb2_grpc.SortedMapMethodsStub(
                    self.channel_manager.get_channel_to(
                        self.placement_client.address_for_actor(
                            headers.application_id,
                            headers.state_ref,
                        )
                    )
                )
                call = stub.Remove(
                    request=request,
                    metadata=headers.to_grpc_metadata(),
                )
                try:
                    return await call
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.UNAVAILABLE:
                        await call_backoff()
                        continue

                    # Reconstitute the error that the server threw, if it was a declared error.
                    status = await IMPORT_rpc_status_async.from_call(call)
                    if status is not None:
                        raise SortedMap.RemoveAborted.from_status(
                            status
                        ) from None
                    raise SortedMap.RemoveAborted.from_grpc_aio_rpc_error(
                        error
                     ) from None

        elif method == 'Get':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method 'Get' is invalid for servicer SortedMap."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=f"Method '{method}' is invalid"
            )
        elif method == 'Range':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method 'Range' is invalid for servicer SortedMap."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=f"Method '{method}' is invalid"
            )
        elif method == 'ReverseRange':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method 'ReverseRange' is invalid for servicer SortedMap."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=f"Method '{method}' is invalid"
            )
        else:
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                f"Method '{method}' is invalid for servicer SortedMap."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )

    async def dispatch(
        self,
        task: IMPORT_reboot_aio_tasks.TaskEffect,
        *,
        only_validate: bool = False,
    ) -> IMPORT_reboot_aio_internals_tasks_dispatcher.TaskResponseOrError:
        """Dispatches the tasks to execute unless 'only_validate' is set to
        true, in which case just ensures that the task actually exists.
        Note that this function will be called *by* tasks_dispatcher; it will
        not itself call into tasks_dispatcher."""

        if 'Insert' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.std.collections.v1.sorted_map_pb2.InsertResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def run_Insert(
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (SortedMapWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Insert(
                            IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.InsertRequest, task.request),
                            bearer_token=__internal_magic_token__,
                            idempotency=IMPORT_reboot_aio_idempotency.Idempotency(
                                alias=f'Task {task.task_id.task_uuid}',
                            ),
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except IMPORT_asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError()))
                            return (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError())
                        else:
                            raise
                    except IMPORT_reboot.aio.aborted.Aborted as aborted:
                        error_type = f'{aborted.error.__class__.__module__}.{aborted.error.__class__.__qualname__}'
                        # Do not retry a task if the error was specified in the
                        # proto file.
                        if error_type in self._specified_errors_by_service_method_name.get('rbt.std.collections.v1.SortedMapMethods.Insert', []):
                            await complete(task, (None, aborted.error))
                            return (None, aborted.error)
                        raise


            return await run_Insert(
                self.create_context(
                    headers=IMPORT_reboot_aio_headers.Headers(
                        application_id=self.application_id,
                        state_ref=IMPORT_reboot_aio_types.StateRef(task.task_id.state_ref),
                    ),
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'Remove' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.std.collections.v1.sorted_map_pb2.RemoveResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def run_Remove(
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (SortedMapWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Remove(
                            IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RemoveRequest, task.request),
                            bearer_token=__internal_magic_token__,
                            idempotency=IMPORT_reboot_aio_idempotency.Idempotency(
                                alias=f'Task {task.task_id.task_uuid}',
                            ),
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except IMPORT_asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError()))
                            return (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError())
                        else:
                            raise
                    except IMPORT_reboot.aio.aborted.Aborted as aborted:
                        error_type = f'{aborted.error.__class__.__module__}.{aborted.error.__class__.__qualname__}'
                        # Do not retry a task if the error was specified in the
                        # proto file.
                        if error_type in self._specified_errors_by_service_method_name.get('rbt.std.collections.v1.SortedMapMethods.Remove', []):
                            await complete(task, (None, aborted.error))
                            return (None, aborted.error)
                        raise


            return await run_Remove(
                self.create_context(
                    headers=IMPORT_reboot_aio_headers.Headers(
                        application_id=self.application_id,
                        state_ref=IMPORT_reboot_aio_types.StateRef(task.task_id.state_ref),
                    ),
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'Get' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.std.collections.v1.sorted_map_pb2.GetResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def run_Get(
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (SortedMapWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Get(
                            IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.GetRequest, task.request),
                            bearer_token=__internal_magic_token__,
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except IMPORT_asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError()))
                            return (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError())
                        else:
                            raise
                    except IMPORT_reboot.aio.aborted.Aborted as aborted:
                        error_type = f'{aborted.error.__class__.__module__}.{aborted.error.__class__.__qualname__}'
                        # Do not retry a task if the error was specified in the
                        # proto file.
                        if error_type in self._specified_errors_by_service_method_name.get('rbt.std.collections.v1.SortedMapMethods.Get', []):
                            await complete(task, (None, aborted.error))
                            return (None, aborted.error)
                        raise


            return await run_Get(
                self.create_context(
                    headers=IMPORT_reboot_aio_headers.Headers(
                        application_id=self.application_id,
                        state_ref=IMPORT_reboot_aio_types.StateRef(task.task_id.state_ref),
                    ),
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'Range' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.std.collections.v1.sorted_map_pb2.RangeResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def run_Range(
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (SortedMapWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Range(
                            IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RangeRequest, task.request),
                            bearer_token=__internal_magic_token__,
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except IMPORT_asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError()))
                            return (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError())
                        else:
                            raise
                    except IMPORT_reboot.aio.aborted.Aborted as aborted:
                        error_type = f'{aborted.error.__class__.__module__}.{aborted.error.__class__.__qualname__}'
                        # Do not retry a task if the error was specified in the
                        # proto file.
                        if error_type in self._specified_errors_by_service_method_name.get('rbt.std.collections.v1.SortedMapMethods.Range', []):
                            await complete(task, (None, aborted.error))
                            return (None, aborted.error)
                        raise


            return await run_Range(
                self.create_context(
                    headers=IMPORT_reboot_aio_headers.Headers(
                        application_id=self.application_id,
                        state_ref=IMPORT_reboot_aio_types.StateRef(task.task_id.state_ref),
                    ),
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'ReverseRange' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def run_ReverseRange(
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (SortedMapWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).ReverseRange(
                            IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest, task.request),
                            bearer_token=__internal_magic_token__,
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except IMPORT_asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError()))
                            return (None, IMPORT_rbt_v1alpha1.tasks_pb2.TaskCancelledError())
                        else:
                            raise
                    except IMPORT_reboot.aio.aborted.Aborted as aborted:
                        error_type = f'{aborted.error.__class__.__module__}.{aborted.error.__class__.__qualname__}'
                        # Do not retry a task if the error was specified in the
                        # proto file.
                        if error_type in self._specified_errors_by_service_method_name.get('rbt.std.collections.v1.SortedMapMethods.ReverseRange', []):
                            await complete(task, (None, aborted.error))
                            return (None, aborted.error)
                        raise


            return await run_ReverseRange(
                self.create_context(
                    headers=IMPORT_reboot_aio_headers.Headers(
                        application_id=self.application_id,
                        state_ref=IMPORT_reboot_aio_types.StateRef(task.task_id.state_ref),
                    ),
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WorkflowContext,
                    task=task,
                )
            )

        # There are no tasks for this service.
        start_or_validate = "start" if not only_validate else "validate"
        raise RuntimeError(
            f"Attempted to {start_or_validate} task '{task.method_name}' "
            f"on 'SortedMap' which does not exist"
        )

    # SortedMap specific methods:
    async def __Insert(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        *,
        validating_effects: bool,
    ) -> SortedMap.InsertEffects:
        try:
            response = (
                await self._servicer.Insert(
                    context=context,
                    state=state,
                    request=request
                )
            )
            IMPORT_reboot_aio_types.assert_type(
                response,
                [
                    rbt.std.collections.v1.sorted_map_pb2.InsertResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, IMPORT_reboot_aio_tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='SortedMap.Insert',
                    validating_effects=validating_effects,
                    context=context,
                )
            return SortedMap.InsertEffects(
                state=state,
                response=response,
                tasks=context._tasks,
                _colocated_upserts=context._colocated_upserts,
            )
        except IMPORT_reboot_aio_contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except IMPORT_reboot_aio_contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except IMPORT_reboot.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `reboot/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: IMPORT_typing.Optional[type] = None
            aborted_type = SortedMap.InsertAborted
            if isinstance(aborted, IMPORT_reboot.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                    logger.warning(
                        f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(IMPORT_traceback.format_exception(aborted))
                    )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                        logger.warning(
                            f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(IMPORT_traceback.format_exception(aborted))
                        )
                    raise IMPORT_reboot.aio.aborted.SystemAborted(
                        IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {aborted}"
                    )

            raise
        except IMPORT_asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except IMPORT_google_protobuf_message.DecodeError as decode_error:
            # We usually see this error when we are trying to construct a proto
            # message which is too deeply nested: protobuf has a limit of 100
            # nested messages. See the limits here:
            #   https://protobuf.dev/programming-guides/proto-limits/

            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {type(decode_error).__name__}: {decode_error}; " +
                    "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                    "See the limits here: https://protobuf.dev/programming-guides/proto-limits/" +
                    ''.join(IMPORT_traceback.format_exception(decode_error))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {decode_error}; "
                        "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
            )
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(IMPORT_traceback.format_exception(exception))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Insert') {type(exception).__name__}: {exception}"
            )

    @IMPORT_reboot_aio_tracing.function_span(
        # We expect an `EffectValidationRetry` exception; that's not an error.
        set_status_on_exception=False
    )
    async def _Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        *,
        validating_effects: bool,
        grpc_context: IMPORT_typing.Optional[IMPORT_grpc.aio.ServicerContext] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = rbt.std.collections.v1.sorted_map_pb2.InsertResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Insert',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False,
            ) as (state, writer):

                effects = await self.__Insert(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )

                await writer.complete(effects)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_Insert(
        self,
        *,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        headers: IMPORT_reboot_aio_headers.Headers,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> tuple[IMPORT_reboot_aio_contexts.WriterContext, rbt.std.collections.v1.sorted_map_pb2.InsertResponse]:
        context: IMPORT_reboot_aio_contexts.WriterContext = self.create_context(
            headers=headers,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.WriterContext,
        )
        response = rbt.std.collections.v1.sorted_map_pb2.InsertResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        IMPORT_reboot_aio_headers.TASK_ID_UUID,
                        str(IMPORT_uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )

            # Try to verify the token if a token verifier exists.
            context.auth = await self._maybe_verify_token(headers)

            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Insert',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                task = SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Insert(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = IMPORT_reboot_aio_state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            IMPORT_reboot_aio_headers.TASK_ID_UUID,
                            str(IMPORT_uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        headers = IMPORT_reboot_aio_headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Fail.
            await grpc_context.abort(
                IMPORT_grpc.StatusCode.UNAVAILABLE,
                f"Consensus '{self.consensus_id}' is not authoritative for this "
                f"request; consensus '{authoritative_consensus}' is.",
            )
            raise  # Unreachable but necessary for mypy.

        @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
            context: IMPORT_typing.Optional[IMPORT_reboot_aio_contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Insert(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WriterContext,
                )
                assert context is not None

                return await self._Insert(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except IMPORT_reboot_aio_contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except IMPORT_reboot.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    IMPORT_rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except IMPORT_asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(IMPORT_traceback.format_exc() if IMPORT_reboot_nodejs_python.should_print_stacktrace() else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __Remove(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        *,
        validating_effects: bool,
    ) -> SortedMap.RemoveEffects:
        try:
            response = (
                await self._servicer.Remove(
                    context=context,
                    state=state,
                    request=request
                )
            )
            IMPORT_reboot_aio_types.assert_type(
                response,
                [
                    rbt.std.collections.v1.sorted_map_pb2.RemoveResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, IMPORT_reboot_aio_tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='SortedMap.Remove',
                    validating_effects=validating_effects,
                    context=context,
                )
            return SortedMap.RemoveEffects(
                state=state,
                response=response,
                tasks=context._tasks,
                _colocated_upserts=context._colocated_upserts,
            )
        except IMPORT_reboot_aio_contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except IMPORT_reboot_aio_contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except IMPORT_reboot.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `reboot/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: IMPORT_typing.Optional[type] = None
            aborted_type = SortedMap.RemoveAborted
            if isinstance(aborted, IMPORT_reboot.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                    logger.warning(
                        f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(IMPORT_traceback.format_exception(aborted))
                    )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                        logger.warning(
                            f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(IMPORT_traceback.format_exception(aborted))
                        )
                    raise IMPORT_reboot.aio.aborted.SystemAborted(
                        IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {aborted}"
                    )

            raise
        except IMPORT_asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except IMPORT_google_protobuf_message.DecodeError as decode_error:
            # We usually see this error when we are trying to construct a proto
            # message which is too deeply nested: protobuf has a limit of 100
            # nested messages. See the limits here:
            #   https://protobuf.dev/programming-guides/proto-limits/

            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {type(decode_error).__name__}: {decode_error}; " +
                    "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                    "See the limits here: https://protobuf.dev/programming-guides/proto-limits/" +
                    ''.join(IMPORT_traceback.format_exception(decode_error))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {decode_error}; "
                        "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
            )
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(IMPORT_traceback.format_exception(exception))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Remove') {type(exception).__name__}: {exception}"
            )

    @IMPORT_reboot_aio_tracing.function_span(
        # We expect an `EffectValidationRetry` exception; that's not an error.
        set_status_on_exception=False
    )
    async def _Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        *,
        validating_effects: bool,
        grpc_context: IMPORT_typing.Optional[IMPORT_grpc.aio.ServicerContext] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = rbt.std.collections.v1.sorted_map_pb2.RemoveResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Remove',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False,
            ) as (state, writer):

                effects = await self.__Remove(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )

                await writer.complete(effects)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_Remove(
        self,
        *,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        headers: IMPORT_reboot_aio_headers.Headers,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> tuple[IMPORT_reboot_aio_contexts.WriterContext, rbt.std.collections.v1.sorted_map_pb2.RemoveResponse]:
        context: IMPORT_reboot_aio_contexts.WriterContext = self.create_context(
            headers=headers,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.WriterContext,
        )
        response = rbt.std.collections.v1.sorted_map_pb2.RemoveResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        IMPORT_reboot_aio_headers.TASK_ID_UUID,
                        str(IMPORT_uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )

            # Try to verify the token if a token verifier exists.
            context.auth = await self._maybe_verify_token(headers)

            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Remove',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                task = SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Remove(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = IMPORT_reboot_aio_state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            IMPORT_reboot_aio_headers.TASK_ID_UUID,
                            str(IMPORT_uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        headers = IMPORT_reboot_aio_headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Fail.
            await grpc_context.abort(
                IMPORT_grpc.StatusCode.UNAVAILABLE,
                f"Consensus '{self.consensus_id}' is not authoritative for this "
                f"request; consensus '{authoritative_consensus}' is.",
            )
            raise  # Unreachable but necessary for mypy.

        @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
            context: IMPORT_typing.Optional[IMPORT_reboot_aio_contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Remove(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.WriterContext,
                )
                assert context is not None

                return await self._Remove(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except IMPORT_reboot_aio_contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except IMPORT_reboot.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    IMPORT_rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except IMPORT_asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(IMPORT_traceback.format_exc() if IMPORT_reboot_nodejs_python.should_print_stacktrace() else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __Get(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        *,
        validating_effects: bool,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        try:
            response = (
                await self._servicer.Get(
                    context=context,
                    state=state,
                    request=request
                )
            )
            IMPORT_reboot_aio_types.assert_type(
                response,
                [
                    rbt.std.collections.v1.sorted_map_pb2.GetResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, IMPORT_reboot_aio_tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='SortedMap.Get',
                    validating_effects=validating_effects,
                    context=context,
                )
            return response
        except IMPORT_reboot_aio_contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except IMPORT_reboot_aio_contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except IMPORT_reboot.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `reboot/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: IMPORT_typing.Optional[type] = None
            aborted_type = SortedMap.GetAborted
            if isinstance(aborted, IMPORT_reboot.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                    logger.warning(
                        f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(IMPORT_traceback.format_exception(aborted))
                    )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                        logger.warning(
                            f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(IMPORT_traceback.format_exception(aborted))
                        )
                    raise IMPORT_reboot.aio.aborted.SystemAborted(
                        IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {aborted}"
                    )

            raise
        except IMPORT_asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except IMPORT_google_protobuf_message.DecodeError as decode_error:
            # We usually see this error when we are trying to construct a proto
            # message which is too deeply nested: protobuf has a limit of 100
            # nested messages. See the limits here:
            #   https://protobuf.dev/programming-guides/proto-limits/

            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {type(decode_error).__name__}: {decode_error}; " +
                    "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                    "See the limits here: https://protobuf.dev/programming-guides/proto-limits/" +
                    ''.join(IMPORT_traceback.format_exception(decode_error))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {decode_error}; "
                        "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
            )
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(IMPORT_traceback.format_exception(exception))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Get') {type(exception).__name__}: {exception}"
            )

    @IMPORT_reboot_aio_tracing.function_span(
        # We expect an `EffectValidationRetry` exception; that's not an error.
        set_status_on_exception=False
    )
    async def _Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        *,
        validating_effects: bool,
        grpc_context: IMPORT_typing.Optional[IMPORT_grpc.aio.ServicerContext] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            authorizer = self._maybe_authorize(
                method_name='rbt.std.collections.v1.SortedMapMethods.Get',
                headers=context._headers,
                auth=context.auth,
                request=request,
            )
            async with self._state_manager.reader(
                context,
                self._servicer.__state_type__,
                authorize=authorizer,
            ) as state:
                if transaction is not None:
                    # We need the transaction to be stored _before_ sending a
                    # response to the user; otherwise they may read two
                    # different states if we crash. Storing the transaction
                    # before even performing the read covers an edge-case:
                    # `StateManager.colocated_read()`, which may be called from
                    # this reader, will read from RocksDB, and will want to use
                    # the transaction.
                    await self._state_manager.transaction_participant_store(
                        transaction
                    )
                response = await self.__Get(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )
                return response

    async def _schedule_Get(
        self,
        *,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        headers: IMPORT_reboot_aio_headers.Headers,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> tuple[IMPORT_reboot_aio_contexts.WriterContext, rbt.std.collections.v1.sorted_map_pb2.GetResponse]:
        context: IMPORT_reboot_aio_contexts.WriterContext = self.create_context(
            headers=headers,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.WriterContext,
        )
        response = rbt.std.collections.v1.sorted_map_pb2.GetResponse()


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )

            # Try to verify the token if a token verifier exists.
            context.auth = await self._maybe_verify_token(headers)

            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Get',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                task = SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Get(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = IMPORT_reboot_aio_state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            IMPORT_reboot_aio_headers.TASK_ID_UUID,
                            str(IMPORT_uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        headers = IMPORT_reboot_aio_headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Fail.
            await grpc_context.abort(
                IMPORT_grpc.StatusCode.UNAVAILABLE,
                f"Consensus '{self.consensus_id}' is not authoritative for this "
                f"request; consensus '{authoritative_consensus}' is.",
            )
            raise  # Unreachable but necessary for mypy.

        @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
            context: IMPORT_typing.Optional[IMPORT_reboot_aio_contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Get(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.ReaderContext,
                )
                assert context is not None

                return await self._Get(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except IMPORT_reboot_aio_contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except IMPORT_reboot.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    IMPORT_rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except IMPORT_asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(IMPORT_traceback.format_exc() if IMPORT_reboot_nodejs_python.should_print_stacktrace() else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __Range(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        *,
        validating_effects: bool,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        try:
            response = (
                await self._servicer.Range(
                    context=context,
                    state=state,
                    request=request
                )
            )
            IMPORT_reboot_aio_types.assert_type(
                response,
                [
                    rbt.std.collections.v1.sorted_map_pb2.RangeResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, IMPORT_reboot_aio_tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='SortedMap.Range',
                    validating_effects=validating_effects,
                    context=context,
                )
            return response
        except IMPORT_reboot_aio_contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except IMPORT_reboot_aio_contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except IMPORT_reboot.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `reboot/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: IMPORT_typing.Optional[type] = None
            aborted_type = SortedMap.RangeAborted
            if isinstance(aborted, IMPORT_reboot.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                    logger.warning(
                        f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(IMPORT_traceback.format_exception(aborted))
                    )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                        logger.warning(
                            f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(IMPORT_traceback.format_exception(aborted))
                        )
                    raise IMPORT_reboot.aio.aborted.SystemAborted(
                        IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {aborted}"
                    )

            raise
        except IMPORT_asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except IMPORT_google_protobuf_message.DecodeError as decode_error:
            # We usually see this error when we are trying to construct a proto
            # message which is too deeply nested: protobuf has a limit of 100
            # nested messages. See the limits here:
            #   https://protobuf.dev/programming-guides/proto-limits/

            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {type(decode_error).__name__}: {decode_error}; " +
                    "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                    "See the limits here: https://protobuf.dev/programming-guides/proto-limits/" +
                    ''.join(IMPORT_traceback.format_exception(decode_error))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {decode_error}; "
                        "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
            )
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(IMPORT_traceback.format_exception(exception))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.Range') {type(exception).__name__}: {exception}"
            )

    @IMPORT_reboot_aio_tracing.function_span(
        # We expect an `EffectValidationRetry` exception; that's not an error.
        set_status_on_exception=False
    )
    async def _Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        *,
        validating_effects: bool,
        grpc_context: IMPORT_typing.Optional[IMPORT_grpc.aio.ServicerContext] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            authorizer = self._maybe_authorize(
                method_name='rbt.std.collections.v1.SortedMapMethods.Range',
                headers=context._headers,
                auth=context.auth,
                request=request,
            )
            async with self._state_manager.reader(
                context,
                self._servicer.__state_type__,
                authorize=authorizer,
            ) as state:
                if transaction is not None:
                    # We need the transaction to be stored _before_ sending a
                    # response to the user; otherwise they may read two
                    # different states if we crash. Storing the transaction
                    # before even performing the read covers an edge-case:
                    # `StateManager.colocated_read()`, which may be called from
                    # this reader, will read from RocksDB, and will want to use
                    # the transaction.
                    await self._state_manager.transaction_participant_store(
                        transaction
                    )
                response = await self.__Range(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )
                return response

    async def _schedule_Range(
        self,
        *,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        headers: IMPORT_reboot_aio_headers.Headers,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> tuple[IMPORT_reboot_aio_contexts.WriterContext, rbt.std.collections.v1.sorted_map_pb2.RangeResponse]:
        context: IMPORT_reboot_aio_contexts.WriterContext = self.create_context(
            headers=headers,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.WriterContext,
        )
        response = rbt.std.collections.v1.sorted_map_pb2.RangeResponse()


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )

            # Try to verify the token if a token verifier exists.
            context.auth = await self._maybe_verify_token(headers)

            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.Range',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                task = SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Range(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = IMPORT_reboot_aio_state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            IMPORT_reboot_aio_headers.TASK_ID_UUID,
                            str(IMPORT_uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        headers = IMPORT_reboot_aio_headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Fail.
            await grpc_context.abort(
                IMPORT_grpc.StatusCode.UNAVAILABLE,
                f"Consensus '{self.consensus_id}' is not authoritative for this "
                f"request; consensus '{authoritative_consensus}' is.",
            )
            raise  # Unreachable but necessary for mypy.

        @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
            context: IMPORT_typing.Optional[IMPORT_reboot_aio_contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Range(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.ReaderContext,
                )
                assert context is not None

                return await self._Range(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except IMPORT_reboot_aio_contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except IMPORT_reboot.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    IMPORT_rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except IMPORT_asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(IMPORT_traceback.format_exc() if IMPORT_reboot_nodejs_python.should_print_stacktrace() else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __ReverseRange(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        *,
        validating_effects: bool,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        try:
            response = (
                await self._servicer.ReverseRange(
                    context=context,
                    state=state,
                    request=request
                )
            )
            IMPORT_reboot_aio_types.assert_type(
                response,
                [
                    rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, IMPORT_reboot_aio_tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='SortedMap.ReverseRange',
                    validating_effects=validating_effects,
                    context=context,
                )
            return response
        except IMPORT_reboot_aio_contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except IMPORT_reboot_aio_contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except IMPORT_reboot.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `reboot/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: IMPORT_typing.Optional[type] = None
            aborted_type = SortedMap.ReverseRangeAborted
            if isinstance(aborted, IMPORT_reboot.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                    logger.warning(
                        f"Unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(IMPORT_traceback.format_exception(aborted))
                    )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                        logger.warning(
                            f"Unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(IMPORT_traceback.format_exception(aborted))
                        )
                    raise IMPORT_reboot.aio.aborted.SystemAborted(
                        IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {aborted}"
                    )

            raise
        except IMPORT_asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except IMPORT_google_protobuf_message.DecodeError as decode_error:
            # We usually see this error when we are trying to construct a proto
            # message which is too deeply nested: protobuf has a limit of 100
            # nested messages. See the limits here:
            #   https://protobuf.dev/programming-guides/proto-limits/

            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {type(decode_error).__name__}: {decode_error}; " +
                    "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                    "See the limits here: https://protobuf.dev/programming-guides/proto-limits/" +
                    ''.join(IMPORT_traceback.format_exception(decode_error))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {decode_error}; "
                        "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
            )
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if IMPORT_reboot_nodejs_python.should_print_stacktrace():
                logger.warning(
                    f"Unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(IMPORT_traceback.format_exception(exception))
                )
            raise IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.std.collections.v1.SortedMap.ReverseRange') {type(exception).__name__}: {exception}"
            )

    @IMPORT_reboot_aio_tracing.function_span(
        # We expect an `EffectValidationRetry` exception; that's not an error.
        set_status_on_exception=False
    )
    async def _ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        *,
        validating_effects: bool,
        grpc_context: IMPORT_typing.Optional[IMPORT_grpc.aio.ServicerContext] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            authorizer = self._maybe_authorize(
                method_name='rbt.std.collections.v1.SortedMapMethods.ReverseRange',
                headers=context._headers,
                auth=context.auth,
                request=request,
            )
            async with self._state_manager.reader(
                context,
                self._servicer.__state_type__,
                authorize=authorizer,
            ) as state:
                if transaction is not None:
                    # We need the transaction to be stored _before_ sending a
                    # response to the user; otherwise they may read two
                    # different states if we crash. Storing the transaction
                    # before even performing the read covers an edge-case:
                    # `StateManager.colocated_read()`, which may be called from
                    # this reader, will read from RocksDB, and will want to use
                    # the transaction.
                    await self._state_manager.transaction_participant_store(
                        transaction
                    )
                response = await self.__ReverseRange(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )
                return response

    async def _schedule_ReverseRange(
        self,
        *,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        headers: IMPORT_reboot_aio_headers.Headers,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> tuple[IMPORT_reboot_aio_contexts.WriterContext, rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse]:
        context: IMPORT_reboot_aio_contexts.WriterContext = self.create_context(
            headers=headers,
            state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            context_type=IMPORT_reboot_aio_contexts.WriterContext,
        )
        response = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse()


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )

            # Try to verify the token if a token verifier exists.
            context.auth = await self._maybe_verify_token(headers)

            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                authorize=self._maybe_authorize(
                    method_name='rbt.std.collections.v1.SortedMapMethods.ReverseRange',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                ),
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                task = SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).ReverseRange(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = IMPORT_reboot_aio_state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            IMPORT_reboot_aio_headers.TASK_ID_UUID,
                            str(IMPORT_uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        grpc_context: IMPORT_grpc.aio.ServicerContext,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        headers = IMPORT_reboot_aio_headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id,
            headers.state_ref,
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Fail.
            await grpc_context.abort(
                IMPORT_grpc.StatusCode.UNAVAILABLE,
                f"Consensus '{self.consensus_id}' is not authoritative for this "
                f"request; consensus '{authoritative_consensus}' is.",
            )
            raise  # Unreachable but necessary for mypy.

        @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
            context: IMPORT_typing.Optional[IMPORT_reboot_aio_contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_ReverseRange(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    context_type=IMPORT_reboot_aio_contexts.ReaderContext,
                )
                assert context is not None

                return await self._ReverseRange(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except IMPORT_reboot_aio_contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except IMPORT_reboot.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    IMPORT_rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except IMPORT_asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(IMPORT_traceback.format_exc() if IMPORT_reboot_nodejs_python.should_print_stacktrace() else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    def _maybe_authorize(
        self,
        *,
        method_name: str,
        headers: IMPORT_reboot_aio_headers.Headers,
        auth: IMPORT_typing.Optional[IMPORT_reboot.aio.auth.Auth],
        request: IMPORT_typing.Optional[SortedMapRequestTypes] = None,
    ) -> IMPORT_typing.Optional[IMPORT_typing.Callable[[IMPORT_typing.Optional[SortedMapStateType]], IMPORT_typing.Awaitable[None]]]:
        """Returns a function to check authorization for the given method.

        Raises `PermissionDenied` in case Authorizer is present but the request
        is not authorized.
        """
        # To authorize internal calls, we use an internal magic token.
        if headers.bearer_token == __internal_magic_token__:
            return None

        assert self._authorizer is not None

        async def authorize(state: IMPORT_typing.Optional[SortedMapStateType]) -> None:
            # Create context for the authorizer. This is a `ReaderContext`
            # independently of the calling context.
            with self.use_context(
                headers=(
                    # Get headers suitable for doing authorization.
                    headers.copy_for_token_verification_and_authorization()
                ),
                state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                context_type=IMPORT_reboot_aio_contexts.ReaderContext,
            ) as context:
                context.auth = auth

                # Get the authorizer decision.
                authorization_decision = await self._authorizer.authorize(
                    method_name=method_name,
                    context=context,
                    state=state,
                    request=request,
                )

            # Enforce correct authorizer decision type.
            try:
                IMPORT_reboot_aio_types.assert_type(
                    authorization_decision,
                    [
                        IMPORT_rbt_v1alpha1.errors_pb2.Ok,
                        IMPORT_rbt_v1alpha1.errors_pb2.Unauthenticated,
                        IMPORT_rbt_v1alpha1.errors_pb2.PermissionDenied,
                    ]
                )
            except TypeError as e:
                # Retyping.cast the exception to provide more context.
                authorizer_type = f"{type(self._authorizer).__module__}.{type(self._authorizer).__name__}"
                raise TypeError(
                    f"Authorizer '{authorizer_type}' "
                    f"returned unexpected type '{type(authorization_decision).__name__}' "
                    f"for method '{method_name}' on "
                    f"`rbt.std.collections.v1.SortedMap('{headers.state_ref.id}')`"
                ) from e

            # If the decision is not `True`, raise a `SystemAborted` with either a
            # `PermissionDenied` error (in case of `False`) or an `Unauthenticated`
            # error.
            if not isinstance(authorization_decision, IMPORT_rbt_v1alpha1.errors_pb2.Ok):
                if isinstance(authorization_decision, IMPORT_rbt_v1alpha1.errors_pb2.Unauthenticated):
                    logger.warning(
                        f"Unauthenticated call to '{method_name}' on "
                        f"`rbt.std.collections.v1.SortedMap('{headers.state_ref.id}')`"
                    )

                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    authorization_decision,
                    message=
                    f"You are not authorized to call '{method_name}' on "
                    f"`rbt.std.collections.v1.SortedMap('{headers.state_ref.id}')`"
                )

        return authorize

    async def _maybe_verify_token(self, headers: IMPORT_reboot_aio_headers.Headers) -> IMPORT_typing.Optional[IMPORT_reboot.aio.auth.Auth]:
        """Verify the bearer token and if a token verifier is present.

        Returns the (optional) `reboot.aio.auth.Auth` object
        produced by the token verifier if the token can be verified.
        """
        if self._token_verifier is not None:
            if headers.bearer_token == __internal_magic_token__:
                return None

            with self.use_context(
                headers=(
                    # Get headers suitable for doing token verification.
                    headers.copy_for_token_verification_and_authorization()
                ),
                state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                context_type=IMPORT_reboot_aio_contexts.ReaderContext,
            ) as context:
                return await self._token_verifier.verify_token(
                    context=context,
                    token=headers.bearer_token,
                )

        return None


############################ Client Stubs ############################
# This section is relevant for clients accessing a Reboot service. Since
# servicers are themselves often clients also, this code is generated for
# them also.


class _SortedMapStub(IMPORT_reboot_aio_stubs.Stub):

    __state_type_name__ = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap')

    def __init__(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        # Within a Reboot context we do not pass on the caller's bearer token, as that might
        # have security implications - we cannot simply trust any service we are calling with
        # the user's credentials. Instead, the developer can rely on the default app-internal
        # auth, or override that and set an explicit bearer token.
        #
        # In the case of `ExternalContext`, however, its `bearer_token` was set specifically
        # by the developer for the purpose of making these calls.
        app_internal_authorization: IMPORT_typing.Optional[str] = None
        if isinstance(context, IMPORT_reboot_aio_external.ExternalContext):
            # Note that only `ExternalContext` even has a `bearer_token` field.
            bearer_token = context.bearer_token
            app_internal_authorization = context.app_internal_authorization

        super().__init__(
            channel_manager=context.channel_manager,
            idempotency_manager=context,
            state_ref=state_ref,
            context=context if isinstance(context, IMPORT_reboot_aio_contexts.Context) else None,
            bearer_token=bearer_token,
            app_internal_authorization=app_internal_authorization,
        )

        # All the channels for all services of this state will go to the same
        # place, so we can just get a single channel and share it across all
        # stubs.
        channel = self._channel_manager.get_channel_to_state(
            self.__state_type_name__, state_ref
        )
        self._rbt_std_collections_v1_sortedmapmethods_stub = rbt.std.collections.v1.sorted_map_pb2_grpc.SortedMapMethodsStub(channel)


class SortedMapReaderStub(_SortedMapStub):

    def __init__(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        *,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.ReaderContext, IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext, IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # SortedMap specific methods:


    async def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Get',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Get,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.GetResponse,
            aborted_type=SortedMap.GetAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Range',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Range,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.RangeResponse,
            aborted_type=SortedMap.RangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'ReverseRange',
            self._rbt_std_collections_v1_sortedmapmethods_stub.ReverseRange,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse,
            aborted_type=SortedMap.ReverseRangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call



class SortedMapWriterStub(_SortedMapStub):

    def __init__(
        self,
        context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        *,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.TransactionContext, IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # SortedMap specific methods:
    async def Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Insert',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Insert',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Insert,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.InsertResponse,
                aborted_type=SortedMap.InsertAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                return await call

    async def Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Remove',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Remove',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Remove,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.RemoveResponse,
                aborted_type=SortedMap.RemoveAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                return await call

    async def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Get',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Get,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.GetResponse,
            aborted_type=SortedMap.GetAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Range',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Range,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.RangeResponse,
            aborted_type=SortedMap.RangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'ReverseRange',
            self._rbt_std_collections_v1_sortedmapmethods_stub.ReverseRange,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse,
            aborted_type=SortedMap.ReverseRangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call


class SortedMapWorkflowStub(_SortedMapStub):

    def __init__(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.TransactionContext, IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # SortedMap specific methods:
    async def Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Insert',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Insert',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Insert,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.InsertResponse,
                aborted_type=SortedMap.InsertAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                return await call

    async def Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Remove',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Remove',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Remove,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.RemoveResponse,
                aborted_type=SortedMap.RemoveAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                return await call

    async def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Get',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Get,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.GetResponse,
            aborted_type=SortedMap.GetAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'Range',
            self._rbt_std_collections_v1_sortedmapmethods_stub.Range,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.RangeResponse,
            aborted_type=SortedMap.RangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call

    async def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        async with self._call(
            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            'ReverseRange',
            self._rbt_std_collections_v1_sortedmapmethods_stub.ReverseRange,
            request,
            unary=True,
            reader=True,
            response_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse,
            aborted_type=SortedMap.ReverseRangeAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, IMPORT_typing.Awaitable)
            return await call



class SortedMapTasksStub(_SortedMapStub):

    def __init__(
        self,
        context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        *,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.TransactionContext, IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # SortedMap specific methods:
    async def Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Insert',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Insert',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Insert,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.InsertResponse,
                aborted_type=SortedMap.InsertAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[misc, attr-defined]
                    if key == IMPORT_reboot_aio_headers.TASK_ID_UUID:
                        return IMPORT_rbt_v1alpha1.tasks_pb2.TaskId(
                            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=IMPORT_uuid.UUID(value).bytes,
                        )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )
    async def Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Remove',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Remove',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Remove,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.RemoveResponse,
                aborted_type=SortedMap.RemoveAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[misc, attr-defined]
                    if key == IMPORT_reboot_aio_headers.TASK_ID_UUID:
                        return IMPORT_rbt_v1alpha1.tasks_pb2.TaskId(
                            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=IMPORT_uuid.UUID(value).bytes,
                        )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )
    async def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Get',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Get',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Get,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.GetResponse,
                aborted_type=SortedMap.GetAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[misc, attr-defined]
                    if key == IMPORT_reboot_aio_headers.TASK_ID_UUID:
                        return IMPORT_rbt_v1alpha1.tasks_pb2.TaskId(
                            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=IMPORT_uuid.UUID(value).bytes,
                        )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )
    async def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='Range',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'Range',
                self._rbt_std_collections_v1_sortedmapmethods_stub.Range,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.RangeResponse,
                aborted_type=SortedMap.RangeAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[misc, attr-defined]
                    if key == IMPORT_reboot_aio_headers.TASK_ID_UUID:
                        return IMPORT_rbt_v1alpha1.tasks_pb2.TaskId(
                            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=IMPORT_uuid.UUID(value).bytes,
                        )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )
    async def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
        *,
        metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
        idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
        with self._idempotency_manager.idempotently(
            state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            service_name=IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
            method='ReverseRange',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
                'ReverseRange',
                self._rbt_std_collections_v1_sortedmapmethods_stub.ReverseRange,
                request,
                unary=True,
                reader=False,
                response_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse,
                aborted_type=SortedMap.ReverseRangeAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, IMPORT_typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[misc, attr-defined]
                    if key == IMPORT_reboot_aio_headers.TASK_ID_UUID:
                        return IMPORT_rbt_v1alpha1.tasks_pb2.TaskId(
                            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=IMPORT_uuid.UUID(value).bytes,
                        )
                raise IMPORT_reboot.aio.aborted.SystemAborted(
                    IMPORT_rbt_v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )


class SortedMapServicerTasksStub(_SortedMapStub):

    _context: IMPORT_reboot_aio_contexts.WriterContext

    def __init__(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        *,
        state_ref: IMPORT_reboot_aio_types.StateRef,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ):
        IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.WriterContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

        self._context = context

    # SortedMap specific methods:
    def Insert(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
        *,
        schedule: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
    ) -> IMPORT_reboot_aio_tasks.TaskEffect:
        schedule = ensure_has_timezone(when=schedule)
        task = IMPORT_reboot_aio_tasks.TaskEffect(
            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            method_name='Insert',
            request=request,
            schedule=(IMPORT_reboot_time_DateTimeWithTimeZone.now() + schedule) if isinstance(
                schedule, IMPORT_datetime_timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def Remove(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
        *,
        schedule: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
    ) -> IMPORT_reboot_aio_tasks.TaskEffect:
        schedule = ensure_has_timezone(when=schedule)
        task = IMPORT_reboot_aio_tasks.TaskEffect(
            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            method_name='Remove',
            request=request,
            schedule=(IMPORT_reboot_time_DateTimeWithTimeZone.now() + schedule) if isinstance(
                schedule, IMPORT_datetime_timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def Get(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
        *,
        schedule: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
    ) -> IMPORT_reboot_aio_tasks.TaskEffect:
        schedule = ensure_has_timezone(when=schedule)
        task = IMPORT_reboot_aio_tasks.TaskEffect(
            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            method_name='Get',
            request=request,
            schedule=(IMPORT_reboot_time_DateTimeWithTimeZone.now() + schedule) if isinstance(
                schedule, IMPORT_datetime_timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def Range(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
        *,
        schedule: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
    ) -> IMPORT_reboot_aio_tasks.TaskEffect:
        schedule = ensure_has_timezone(when=schedule)
        task = IMPORT_reboot_aio_tasks.TaskEffect(
            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            method_name='Range',
            request=request,
            schedule=(IMPORT_reboot_time_DateTimeWithTimeZone.now() + schedule) if isinstance(
                schedule, IMPORT_datetime_timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def ReverseRange(
        self,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
        *,
        schedule: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
    ) -> IMPORT_reboot_aio_tasks.TaskEffect:
        schedule = ensure_has_timezone(when=schedule)
        task = IMPORT_reboot_aio_tasks.TaskEffect(
            state_type=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
            state_ref=self._headers.state_ref,
            method_name='ReverseRange',
            request=request,
            schedule=(IMPORT_reboot_time_DateTimeWithTimeZone.now() + schedule) if isinstance(
                schedule, IMPORT_datetime_timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task



############################ Authorizers ############################
# Relevant to servicers; irrelevant to clients.

SortedMapStateType: IMPORT_typing.TypeAlias = rbt.std.collections.v1.sorted_map_pb2.SortedMap
SortedMapRequestTypes: IMPORT_typing.TypeAlias = \
        rbt.std.collections.v1.sorted_map_pb2.InsertRequest \
        | rbt.std.collections.v1.sorted_map_pb2.RemoveRequest \
        | rbt.std.collections.v1.sorted_map_pb2.GetRequest \
        | rbt.std.collections.v1.sorted_map_pb2.RangeRequest \
        | rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest

class SortedMapAuthorizer(
    IMPORT_reboot.aio.auth.authorizers.Authorizer[SortedMapStateType, SortedMapRequestTypes],
):
    StateType: IMPORT_typing.TypeAlias = SortedMapStateType
    RequestTypes: IMPORT_typing.TypeAlias = SortedMapRequestTypes

    def __init__(
        self,
        *,
        Insert: IMPORT_typing.Optional[
            IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
              rbt.std.collections.v1.sorted_map_pb2.SortedMap,
              rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
            ]
        ] = None,
        Remove: IMPORT_typing.Optional[
            IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
              rbt.std.collections.v1.sorted_map_pb2.SortedMap,
              rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
            ]
        ] = None,
        Get: IMPORT_typing.Optional[
            IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
              rbt.std.collections.v1.sorted_map_pb2.SortedMap,
              rbt.std.collections.v1.sorted_map_pb2.GetRequest,
            ]
        ] = None,
        Range: IMPORT_typing.Optional[
            IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
              rbt.std.collections.v1.sorted_map_pb2.SortedMap,
              rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
            ]
        ] = None,
        ReverseRange: IMPORT_typing.Optional[
            IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
              rbt.std.collections.v1.sorted_map_pb2.SortedMap,
              rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
            ]
        ] = None,
        # NOTE: using `_` prefix for `_default` so as not to collide
        # with any method names since a prefixed `_` is forbidden by
        # our protoc plugins.
        _default: IMPORT_reboot.aio.auth.authorizers.AuthorizerRule[
            rbt.std.collections.v1.sorted_map_pb2.SortedMap,
            IMPORT_google_protobuf_message.Message,
        ] = IMPORT_reboot.aio.auth.authorizers.allow_if(
            all=[IMPORT_reboot.aio.auth.authorizers.is_app_internal],
        ),
    ):
        self._Insert = Insert
        self._Remove = Remove
        self._Get = Get
        self._Range = Range
        self._ReverseRange = ReverseRange
        self.__default = _default

    async def authorize(
        self,
        *,
        method_name: str,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: IMPORT_typing.Optional[SortedMapStateType],
        request: IMPORT_typing.Optional[SortedMapRequestTypes],
        **kwargs,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        if method_name == 'rbt.std.collections.v1.SortedMapMethods.Insert':
            return await self.Insert(
                context=context,
                state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
                request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.InsertRequest, request),
            )
        elif method_name == 'rbt.std.collections.v1.SortedMapMethods.Remove':
            return await self.Remove(
                context=context,
                state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
                request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RemoveRequest, request),
            )
        elif method_name == 'rbt.std.collections.v1.SortedMapMethods.Get':
            return await self.Get(
                context=context,
                state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
                request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.GetRequest, request),
            )
        elif method_name == 'rbt.std.collections.v1.SortedMapMethods.Range':
            return await self.Range(
                context=context,
                state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
                request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RangeRequest, request),
            )
        elif method_name == 'rbt.std.collections.v1.SortedMapMethods.ReverseRange':
            return await self.ReverseRange(
                context=context,
                state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
                request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest, request),
            )
        else:
            return IMPORT_rbt_v1alpha1.errors_pb2.PermissionDenied()

    # For 'rbt.std.collections.v1.SortedMapMethods.Insert'.
    async def Insert(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        return await (self._Insert or self.__default).execute(
            context=context,
            state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
            request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.InsertRequest, request),
        )

    # For 'rbt.std.collections.v1.SortedMapMethods.Remove'.
    async def Remove(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        return await (self._Remove or self.__default).execute(
            context=context,
            state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
            request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RemoveRequest, request),
        )

    # For 'rbt.std.collections.v1.SortedMapMethods.Get'.
    async def Get(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        return await (self._Get or self.__default).execute(
            context=context,
            state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
            request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.GetRequest, request),
        )

    # For 'rbt.std.collections.v1.SortedMapMethods.Range'.
    async def Range(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        return await (self._Range or self.__default).execute(
            context=context,
            state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
            request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.RangeRequest, request),
        )

    # For 'rbt.std.collections.v1.SortedMapMethods.ReverseRange'.
    async def ReverseRange(
        self,
        *,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
    ) -> IMPORT_reboot.aio.auth.authorizers.Authorizer.Decision:
        return await (self._ReverseRange or self.__default).execute(
            context=context,
            state=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.SortedMap, state),
            request=IMPORT_typing.cast(rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest, request),
        )



############################ Reboot Servicers ############################
# Base classes for server-side implementations of Reboot servicers.
# Irrelevant to clients.

class SortedMapServicer(IMPORT_reboot_aio_servicers.Servicer):
    Authorizer: IMPORT_typing.TypeAlias = SortedMapAuthorizer

    __service_names__ = [
        IMPORT_reboot_aio_types.ServiceName('rbt.std.collections.v1.SortedMapMethods'),
    ]
    __state_type_name__ = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap')
    __state_type__ = rbt.std.collections.v1.sorted_map_pb2.SortedMap
    __file_descriptor__ = rbt.std.collections.v1.sorted_map_pb2.DESCRIPTOR

    def __init__(self):
        super().__init__()
        # NOTE: need to hold on to the middleware so we can do inline
        # writes (see 'self.write(...)').
        #
        # Because '_middleware' is not really private this does mean
        # users may do possibly dangerous things, but this is no more
        # likely given they could have already overridden
        # 'create_middleware()'.
        self._middleware: IMPORT_typing.Optional[SortedMapServicerMiddleware] = None

    def create_middleware(
        self,
        *,
        application_id: IMPORT_reboot_aio_types.ApplicationId,
        consensus_id: IMPORT_reboot_aio_types.ConsensusId,
        state_manager: IMPORT_reboot_aio_state_managers.StateManager,
        placement_client: IMPORT_reboot_aio_placement.PlacementClient,
        channel_manager: IMPORT_reboot_aio_internals_channel_manager._ChannelManager,
        tasks_cache: IMPORT_reboot_aio_internals_tasks_cache.TasksCache,
        token_verifier: IMPORT_typing.Optional[IMPORT_reboot.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: IMPORT_reboot_aio_contexts.EffectValidation,
        app_internal_api_key_secret: str,
        ready: IMPORT_asyncio.Event,
    ) -> SortedMapServicerMiddleware:
        self._middleware = SortedMapServicerMiddleware(
            servicer=self,
            application_id=application_id,
            consensus_id=consensus_id,
            state_manager=state_manager,
            placement_client=placement_client,
            channel_manager=channel_manager,
            tasks_cache=tasks_cache,
            token_verifier=token_verifier,
            effect_validation=effect_validation,
            app_internal_api_key_secret=app_internal_api_key_secret,
            ready=ready,
        )
        return self._middleware

    def authorizer(self) -> IMPORT_typing.Optional[IMPORT_reboot.aio.auth.authorizers.Authorizer | IMPORT_reboot.aio.auth.authorizers.AuthorizerRule]:
        return None

    def token_verifier(self) -> IMPORT_typing.Optional[IMPORT_reboot.aio.auth.token_verifiers.TokenVerifier]:
        return None

    def ref(
        self,
        *,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> SortedMap.WeakReference[SortedMap.WeakReference._WriterSchedule]:
        context = IMPORT_reboot_aio_contexts.Context.get()

        if context is None:
            raise RuntimeError(
                'Missing asyncio context variable `context`; '
                'are you using this class without Reboot?'
            )

        return SortedMap.WeakReference(
            # TODO(https://github.com/reboot-dev/mono/issues/3226): add support for calling other applications.
            # For now this always stays within the application that creates the context.
            application_id=None,
            state_id=context._state_ref.id,
            schedule_type=SortedMap.WeakReference._WriterSchedule,
            # If the user didn't specify a bearer token we may still end up using the app-internal bearer token,
            # but that's decided at the time of the call.
            bearer_token=bearer_token,
        )

    class Effects(IMPORT_reboot_aio_state_managers.Effects):
        def __init__(
            self,
            *,
            state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
            response: IMPORT_typing.Optional[IMPORT_google_protobuf_message.Message] = None,
            tasks: IMPORT_typing.Optional[list[IMPORT_reboot_aio_tasks.TaskEffect]] = None,
            _colocated_upserts: IMPORT_typing.Optional[list[tuple[str, IMPORT_typing.Optional[bytes]]]] = None,
        ):
            IMPORT_reboot_aio_types.assert_type(state, [rbt.std.collections.v1.sorted_map_pb2.SortedMap])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)

    # For 'rbt.std.collections.v1.SortedMapMethods.Insert'.
    class InsertEffects(Effects):
        def __init__(
            self,
            *,
            state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
            response: rbt.std.collections.v1.sorted_map_pb2.InsertResponse,
            tasks: IMPORT_typing.Optional[list[IMPORT_reboot_aio_tasks.TaskEffect]] = None,
            _colocated_upserts: IMPORT_typing.Optional[list[tuple[str, IMPORT_typing.Optional[bytes]]]] = None,
        ):
            IMPORT_reboot_aio_types.assert_type(state, [rbt.std.collections.v1.sorted_map_pb2.SortedMap])
            IMPORT_reboot_aio_types.assert_type(response, [rbt.std.collections.v1.sorted_map_pb2.InsertResponse])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)


    # For 'rbt.std.collections.v1.SortedMapMethods.Remove'.
    class RemoveEffects(Effects):
        def __init__(
            self,
            *,
            state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
            response: rbt.std.collections.v1.sorted_map_pb2.RemoveResponse,
            tasks: IMPORT_typing.Optional[list[IMPORT_reboot_aio_tasks.TaskEffect]] = None,
            _colocated_upserts: IMPORT_typing.Optional[list[tuple[str, IMPORT_typing.Optional[bytes]]]] = None,
        ):
            IMPORT_reboot_aio_types.assert_type(state, [rbt.std.collections.v1.sorted_map_pb2.SortedMap])
            IMPORT_reboot_aio_types.assert_type(response, [rbt.std.collections.v1.sorted_map_pb2.RemoveResponse])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)






    InlineWriterCallableResult = IMPORT_typing.TypeVar('InlineWriterCallableResult', covariant=True)

    class InlineWriterCallable(IMPORT_typing.Protocol[InlineWriterCallableResult]):
        async def __call__(
            self,
            state: rbt.std.collections.v1.sorted_map_pb2.SortedMap
        ) -> SortedMapServicer.InlineWriterCallableResult:
            ...

    class _State:

        def __init__(
            self,
            servicer,
        ):
            self._servicer = servicer

        async def read(
            self, context: IMPORT_reboot_aio_contexts.WorkflowContext
        ) -> rbt.std.collections.v1.sorted_map_pb2.SortedMap:
            """Read the current state within a workflow."""

            IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.WorkflowContext])

            if self._servicer._middleware is None:
                raise RuntimeError(
                    'Reboot middleware was not created; '
                    'are you using this class without Reboot?'
                )

            return await self._servicer._middleware._state_manager.read(
                context, self._servicer.__state_type__
            )

        @IMPORT_typing.overload
        async def write(
            self,
            idempotency_alias: str,
            context: IMPORT_reboot_aio_contexts.WorkflowContext,
            writer: SortedMapServicer.InlineWriterCallable[None],
            __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
            *,
            type: type = type(None),
        ) -> None:
            ...

        @IMPORT_typing.overload
        async def write(
            self,
            idempotency_alias: str,
            context: IMPORT_reboot_aio_contexts.WorkflowContext,
            writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
            __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
            *,
            type: type[SortedMapServicer.InlineWriterCallableResult],
        ) -> SortedMapServicer.InlineWriterCallableResult:
            ...

        async def write(
            self,
            idempotency_alias: str,
            context: IMPORT_reboot_aio_contexts.WorkflowContext,
            writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
            __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
            *,
            type: type = type(None),
        ) -> SortedMapServicer.InlineWriterCallableResult:
            """Perform an "inline write" within a workflow."""
            return await self.idempotently(idempotency_alias).write(
                context, writer, __options__, type=type
            )

        class _Idempotently:

            def __init__(
                self,
                *,
                servicer: SortedMapServicer,
                idempotency: IMPORT_reboot_aio_idempotency.Idempotency,
            ):
                self._servicer = servicer
                self._idempotency = idempotency

            @IMPORT_typing.overload
            async def write(
                self,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[None],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                type: type = type(None),
                check_type: bool = True,
            ) -> None:
                ...

            @IMPORT_typing.overload
            async def write(
                self,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                type: type[SortedMapServicer.InlineWriterCallableResult],
                check_type: bool = True,
            ) -> SortedMapServicer.InlineWriterCallableResult:
                ...

            async def write(
                self,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                type: type = type(None),
                check_type: bool = True,
            ) -> SortedMapServicer.InlineWriterCallableResult:
                return await self._write(
                    context,
                    writer,
                    __options__,
                    type_result=type,
                    check_type=check_type,
                    unidempotently=False,
                )

            async def _write(
                self,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                type_result: type,
                check_type: bool,
                unidempotently: bool,
            ) -> SortedMapServicer.InlineWriterCallableResult:
                return await self._write_validating_effects(
                    self._servicer,
                    self._idempotency,
                    context,
                    writer,
                    __options__,
                    type_result=type_result,
                    check_type=check_type,
                    unidempotently=unidempotently,
                )

            @staticmethod
            @IMPORT_reboot_aio_internals_middleware.maybe_run_function_twice_to_validate_effects
            async def _write_validating_effects(
                validating_effects: bool,
                servicer: SortedMapServicer,
                idempotency: IMPORT_reboot_aio_idempotency.Idempotency,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                type_result: type,
                check_type: bool,
                unidempotently: bool,
            ) -> SortedMapServicer.InlineWriterCallableResult:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.idempotency is not None:
                    raise RuntimeError(
                        'Found redundant idempotency in `Options`'
                    )

                assert not idempotency.generate, (
                    "Inline writers are always required to specify either an "
                    "idempotency alias or an idempotency key"
                )

                IMPORT_reboot_aio_types.assert_type(context, [IMPORT_reboot_aio_contexts.WorkflowContext])

                if servicer._middleware is None:
                    raise RuntimeError(
                        'Reboot middleware was not created; '
                        'are you using this class without Reboot?'
                    )

                metadata: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None

                if __options__ is not None:
                    if __options__.metadata is not None:
                        metadata = __options__.metadata

                if metadata is None:
                    metadata = ()

                headers = IMPORT_reboot_aio_headers.Headers(
                    application_id=context.application_id,
                    state_ref=context._state_ref,
                )

                metadata += headers.to_grpc_metadata()

                idempotency_key: IMPORT_typing.Optional[IMPORT_uuid.UUID]
                with context.idempotently(
                    state_type_name=IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                    state_ref=context._state_ref,
                    service_name=None,  # Indicates an inline writer.
                    method=None,  # Indicates an inline writer.
                    request=None,
                    metadata=metadata,
                    idempotency=idempotency,
                ) as idempotency_key:

                    if any(t[0] == IMPORT_reboot_aio_headers.IDEMPOTENCY_KEY_HEADER for t in metadata):
                        raise ValueError(
                            f"Do not set '{IMPORT_reboot_aio_headers.IDEMPOTENCY_KEY_HEADER}' metadata yourself"
                        )

                    if idempotency_key is not None:
                        metadata += (
                            (IMPORT_reboot_aio_headers.IDEMPOTENCY_KEY_HEADER, str(idempotency_key)),
                        )

                    with servicer._middleware.use_context(
                        headers=IMPORT_reboot_aio_headers.Headers.from_grpc_metadata(metadata),
                        state_type_name = IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                        context_type=IMPORT_reboot_aio_contexts.WriterContext,
                    ) as writer_context:
                        # Check if we already have performed this mutation!
                        #
                        # We do this _before_ calling 'transactionally()' because
                        # if this call is for a transaction method _and_ we've
                        # already performed the transaction then we don't want to
                        # become a transaction participant (again) we just want to
                        # return the transaction's response.
                        idempotent_mutation = (
                            servicer._middleware._state_manager.check_for_idempotent_mutation(
                                writer_context
                            )
                        )

                        if idempotent_mutation is not None:
                            assert len(idempotent_mutation.response) != 0
                            response = IMPORT_google_protobuf_wrappers_pb2.BytesValue()
                            response.ParseFromString(idempotent_mutation.response)
                            result: SortedMapServicer.InlineWriterCallableResult = IMPORT_pickle.loads(response.value)

                            if check_type and type(result) is not type_result:
                                raise TypeError(
                                    f"Stored result of type '{type(result).__name__}' from 'writer' "
                                    f"is not of expected type '{type_result.__name__}'; have you changed "
                                    "the 'type' that you expect after having stored a result?"
                                )

                            return result

                        async with servicer._middleware._state_manager.transactionally(
                            writer_context, servicer._middleware.tasks_dispatcher
                        ) as transaction:
                            async with servicer._middleware._state_manager.writer(
                                writer_context,
                                servicer.__state_type__,
                                servicer._middleware.tasks_dispatcher,
                                # TODO: Decide if we want to do any kind of authorization for inline
                                # writers otherwise passing `None` here is fine.
                                authorize=None,
                                transaction=transaction,
                            ) as (state, state_manager_writer):
                                result = await writer(state=state)

                                if check_type and type(result) is not type_result:
                                    raise TypeError(
                                        f"Result of type '{type(result).__name__}' from 'writer' is "
                                        f"not of expected type '{type_result.__name__}'; "
                                        "did you specify an incorrect 'type'?"
                                    )

                                effects = SortedMapServicer.Effects(
                                    state=state,
                                    tasks=context._tasks,
                                    response=IMPORT_google_protobuf_wrappers_pb2.BytesValue(
                                        value=IMPORT_pickle.dumps(result)
                                    ),
                                )

                                task: IMPORT_typing.Optional[IMPORT_reboot_aio_tasks.TaskEffect] = context.task

                                assert task is not None, (
                                    "Should always have a task when running a `workflow`"
                                )

                                method_name = f"SortedMap.{task.method_name} "

                                if unidempotently:
                                    method_name += "unidempotent inline writer"
                                elif idempotency.alias is not None:
                                    method_name += "inline writer with idempotency alias '" + idempotency.alias + "'"
                                else:
                                    assert idempotency.key is not None
                                    method_name += "inline writer with idempotency key=" + str(idempotency.key)

                                servicer._middleware.maybe_raise_effect_validation_retry(
                                    logger=logger,
                                    idempotency_manager=context,
                                    method_name=method_name,
                                    validating_effects=validating_effects,
                                    context=context,
                                )

                                await state_manager_writer.complete(effects)

                                return result

        @IMPORT_typing.overload
        def idempotently(self, alias: str, *, each_iteration: bool = False) -> SortedMapServicer._State._Idempotently:
            ...

        @IMPORT_typing.overload
        def idempotently(self, *, key: IMPORT_uuid.UUID) -> SortedMapServicer._State._Idempotently:
            ...

        def idempotently(
            self,
            alias: IMPORT_typing.Optional[str] = None,
            *,
            key: IMPORT_typing.Optional[IMPORT_uuid.UUID] = None,
            each_iteration: IMPORT_typing.Optional[bool] = None,
        ) -> SortedMapServicer._State._Idempotently:
            if alias is None and key is None:
                raise ValueError(
                    'Inline writers require either an idempotency alias or key'
                )
            return SortedMapServicer._State._Idempotently(
                servicer=self._servicer,
                idempotency=IMPORT_reboot_aio_contexts.Context.idempotency(
                    key=key, alias=alias, each_iteration=each_iteration
                ),
            )

        class _Unidempotently:

            def __init__(
                self,
                *,
                servicer: SortedMapServicer,
            ):
                self._servicer = servicer

            async def write(
                self,
                context: IMPORT_reboot_aio_contexts.WorkflowContext,
                writer: SortedMapServicer.InlineWriterCallable[SortedMapServicer.InlineWriterCallableResult],
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
            ) -> SortedMapServicer.InlineWriterCallableResult:
                return await SortedMapServicer._State._Idempotently(
                    servicer=self._servicer,
                    idempotency=IMPORT_reboot_aio_contexts.Context.idempotency(
                        key=IMPORT_uuid.uuid4()
                    ),
                )._write(
                    context,
                    writer,
                    __options__,
                    type_result=type(None),
                    check_type=False,
                    unidempotently=True,
                )

        def unidempotently(self):
            return SortedMapServicer._State._Unidempotently(
                servicer=self._servicer,
            )

    @property
    def state(self):
        return SortedMapServicer._State(
            servicer=self
        )

    # For 'rbt.std.collections.v1.SortedMapMethods.Insert'.
    @IMPORT_abc_abstractmethod
    async def Insert(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        raise NotImplementedError

    # For 'rbt.std.collections.v1.SortedMapMethods.Remove'.
    @IMPORT_abc_abstractmethod
    async def Remove(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        raise NotImplementedError

    # For 'rbt.std.collections.v1.SortedMapMethods.Get'.
    @IMPORT_abc_abstractmethod
    async def Get(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        raise NotImplementedError

    # For 'rbt.std.collections.v1.SortedMapMethods.Range'.
    @IMPORT_abc_abstractmethod
    async def Range(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        raise NotImplementedError

    # For 'rbt.std.collections.v1.SortedMapMethods.ReverseRange'.
    @IMPORT_abc_abstractmethod
    async def ReverseRange(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        raise NotImplementedError



############################ Clients ############################
# The main developer-facing entrypoints for any Reboot type. Relevant to both
# clients and servicers (who use it to find the right servicer base types, as well
# as often being clients themselves).

# Attach an explicit time time zone to "naive" `datetime` objects. A "naive" `datetime` doesn't have a
# time zone. Such objects are typically interpreted as representing local time, but could be confused
# for objects representing UTC. This helper function disambiguates by explicitly attaching the local
# time zone to `datetime` objects that don't already have an explicit time zone. If the `datetime` object
# is already timezone-aware, we still convert it to our custom `DateTimeWithTimeZone` type.
def ensure_has_timezone(
    *,
    when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
) -> IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone | IMPORT_datetime_timedelta]:
    if isinstance(when, IMPORT_datetime_datetime):
        return IMPORT_reboot_time_DateTimeWithTimeZone.from_datetime(when)
    return when

SortedMap_ScheduleTypeVar = IMPORT_typing.TypeVar('SortedMap_ScheduleTypeVar', 'SortedMap.WeakReference._Schedule', 'SortedMap.WeakReference._WriterSchedule')
SortedMap_IdempotentlyScheduleTypeVar = IMPORT_typing.TypeVar('SortedMap_IdempotentlyScheduleTypeVar', 'SortedMap.WeakReference._Schedule', 'SortedMap.WeakReference._WriterSchedule')

class SortedMap:

    Servicer: IMPORT_typing.TypeAlias = SortedMapServicer

    Effects: IMPORT_typing.TypeAlias = SortedMapServicer.Effects

    Authorizer: IMPORT_typing.TypeAlias = SortedMapAuthorizer

    State: IMPORT_typing.TypeAlias = rbt.std.collections.v1.sorted_map_pb2.SortedMap

    __state_type_name__ = IMPORT_reboot_aio_types.StateTypeName("rbt.std.collections.v1.SortedMap")

    class InsertTask:
        """Represents a scheduled task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be).
        """

        @classmethod
        def retrieve(
            cls,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ):
            return cls(context, task_id=task_id)

        def __init__(
            self,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ) -> None:
            # Depending on the context type (inside or outside a Reboot application)
            # we may or may not know the application ID. If we don't know it, then
            # the `ExternalContext.gateway` will determine it.
            #
            # TODO: in the future we expect to support cross-application calls, in
            #       which case the developer may explicitly pass in an application ID
            #       here.
            self._application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId] = None
            if isinstance(context, IMPORT_reboot_aio_contexts.Context):
                self._application_id = context.application_id
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> IMPORT_typing.Generator[None, None, rbt.std.collections.v1.sorted_map_pb2.InsertResponse]:
            """Awaits for task to finish and returns it's response."""
            async def wait_for_task() -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
                channel = self._channel_manager.get_channel_to_state(
                    self._task_id.state_type,
                    IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                )

                stub = IMPORT_rbt_v1alpha1.tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        IMPORT_rbt_v1alpha1.tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=IMPORT_reboot_aio_headers.Headers(
                            state_ref=IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                            application_id=self._application_id,
                        ).to_grpc_metadata(),
                    )
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.NOT_FOUND:
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise IMPORT_reboot.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: IMPORT_typing.Optional[IMPORT_google_protobuf_any_pb2.Any] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    assert response_or_error is not None
                    assert response_or_error.TypeName() != ""

                    response = rbt.std.collections.v1.sorted_map_pb2.InsertResponse()

                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(IMPORT_uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.std.collections.v1.sorted_map_pb2.InsertResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise IMPORT_asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()

    InsertEffects: IMPORT_typing.TypeAlias = Servicer.InsertEffects

    class InsertAborted(IMPORT_reboot.aio.aborted.Aborted):


        Error = IMPORT_typing.Union[
            IMPORT_reboot.aio.aborted.GrpcError,
            IMPORT_reboot.aio.aborted.RebootError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.REBOOT_ERROR_TYPES
        )

        _error: Error
        _code: IMPORT_grpc.StatusCode
        _message: IMPORT_typing.Optional[str]

        def __init__(
            self,
            error:  IMPORT_reboot.aio.aborted.GrpcError,
            *,
            message: IMPORT_typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            IMPORT_reboot_aio_types.assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Reboot specific or declared method error.
                code = IMPORT_grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> IMPORT_grpc.StatusCode:
            return self._code

        @property
        def message(self) -> IMPORT_typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: IMPORT_google_rpc_status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: IMPORT_grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: IMPORT_google_protobuf_message.Message) -> bool:
            return False

    class RemoveTask:
        """Represents a scheduled task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be).
        """

        @classmethod
        def retrieve(
            cls,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ):
            return cls(context, task_id=task_id)

        def __init__(
            self,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ) -> None:
            # Depending on the context type (inside or outside a Reboot application)
            # we may or may not know the application ID. If we don't know it, then
            # the `ExternalContext.gateway` will determine it.
            #
            # TODO: in the future we expect to support cross-application calls, in
            #       which case the developer may explicitly pass in an application ID
            #       here.
            self._application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId] = None
            if isinstance(context, IMPORT_reboot_aio_contexts.Context):
                self._application_id = context.application_id
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> IMPORT_typing.Generator[None, None, rbt.std.collections.v1.sorted_map_pb2.RemoveResponse]:
            """Awaits for task to finish and returns it's response."""
            async def wait_for_task() -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
                channel = self._channel_manager.get_channel_to_state(
                    self._task_id.state_type,
                    IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                )

                stub = IMPORT_rbt_v1alpha1.tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        IMPORT_rbt_v1alpha1.tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=IMPORT_reboot_aio_headers.Headers(
                            state_ref=IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                            application_id=self._application_id,
                        ).to_grpc_metadata(),
                    )
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.NOT_FOUND:
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise IMPORT_reboot.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: IMPORT_typing.Optional[IMPORT_google_protobuf_any_pb2.Any] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    assert response_or_error is not None
                    assert response_or_error.TypeName() != ""

                    response = rbt.std.collections.v1.sorted_map_pb2.RemoveResponse()

                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(IMPORT_uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.std.collections.v1.sorted_map_pb2.RemoveResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise IMPORT_asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()

    RemoveEffects: IMPORT_typing.TypeAlias = Servicer.RemoveEffects

    class RemoveAborted(IMPORT_reboot.aio.aborted.Aborted):


        Error = IMPORT_typing.Union[
            IMPORT_reboot.aio.aborted.GrpcError,
            IMPORT_reboot.aio.aborted.RebootError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.REBOOT_ERROR_TYPES
        )

        _error: Error
        _code: IMPORT_grpc.StatusCode
        _message: IMPORT_typing.Optional[str]

        def __init__(
            self,
            error:  IMPORT_reboot.aio.aborted.GrpcError,
            *,
            message: IMPORT_typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            IMPORT_reboot_aio_types.assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Reboot specific or declared method error.
                code = IMPORT_grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> IMPORT_grpc.StatusCode:
            return self._code

        @property
        def message(self) -> IMPORT_typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: IMPORT_google_rpc_status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: IMPORT_grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: IMPORT_google_protobuf_message.Message) -> bool:
            return False

    class GetTask:
        """Represents a scheduled task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be).
        """

        @classmethod
        def retrieve(
            cls,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ):
            return cls(context, task_id=task_id)

        def __init__(
            self,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ) -> None:
            # Depending on the context type (inside or outside a Reboot application)
            # we may or may not know the application ID. If we don't know it, then
            # the `ExternalContext.gateway` will determine it.
            #
            # TODO: in the future we expect to support cross-application calls, in
            #       which case the developer may explicitly pass in an application ID
            #       here.
            self._application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId] = None
            if isinstance(context, IMPORT_reboot_aio_contexts.Context):
                self._application_id = context.application_id
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> IMPORT_typing.Generator[None, None, rbt.std.collections.v1.sorted_map_pb2.GetResponse]:
            """Awaits for task to finish and returns it's response."""
            async def wait_for_task() -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
                channel = self._channel_manager.get_channel_to_state(
                    self._task_id.state_type,
                    IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                )

                stub = IMPORT_rbt_v1alpha1.tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        IMPORT_rbt_v1alpha1.tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=IMPORT_reboot_aio_headers.Headers(
                            state_ref=IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                            application_id=self._application_id,
                        ).to_grpc_metadata(),
                    )
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.NOT_FOUND:
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise IMPORT_reboot.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: IMPORT_typing.Optional[IMPORT_google_protobuf_any_pb2.Any] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    assert response_or_error is not None
                    assert response_or_error.TypeName() != ""

                    response = rbt.std.collections.v1.sorted_map_pb2.GetResponse()

                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(IMPORT_uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.std.collections.v1.sorted_map_pb2.GetResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise IMPORT_asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()


    class GetAborted(IMPORT_reboot.aio.aborted.Aborted):


        Error = IMPORT_typing.Union[
            IMPORT_reboot.aio.aborted.GrpcError,
            IMPORT_reboot.aio.aborted.RebootError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.REBOOT_ERROR_TYPES
        )

        _error: Error
        _code: IMPORT_grpc.StatusCode
        _message: IMPORT_typing.Optional[str]

        def __init__(
            self,
            error:  IMPORT_reboot.aio.aborted.GrpcError,
            *,
            message: IMPORT_typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            IMPORT_reboot_aio_types.assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Reboot specific or declared method error.
                code = IMPORT_grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> IMPORT_grpc.StatusCode:
            return self._code

        @property
        def message(self) -> IMPORT_typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: IMPORT_google_rpc_status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: IMPORT_grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: IMPORT_google_protobuf_message.Message) -> bool:
            return False

    class RangeTask:
        """Represents a scheduled task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be).
        """

        @classmethod
        def retrieve(
            cls,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ):
            return cls(context, task_id=task_id)

        def __init__(
            self,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ) -> None:
            # Depending on the context type (inside or outside a Reboot application)
            # we may or may not know the application ID. If we don't know it, then
            # the `ExternalContext.gateway` will determine it.
            #
            # TODO: in the future we expect to support cross-application calls, in
            #       which case the developer may explicitly pass in an application ID
            #       here.
            self._application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId] = None
            if isinstance(context, IMPORT_reboot_aio_contexts.Context):
                self._application_id = context.application_id
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> IMPORT_typing.Generator[None, None, rbt.std.collections.v1.sorted_map_pb2.RangeResponse]:
            """Awaits for task to finish and returns it's response."""
            async def wait_for_task() -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
                channel = self._channel_manager.get_channel_to_state(
                    self._task_id.state_type,
                    IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                )

                stub = IMPORT_rbt_v1alpha1.tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        IMPORT_rbt_v1alpha1.tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=IMPORT_reboot_aio_headers.Headers(
                            state_ref=IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                            application_id=self._application_id,
                        ).to_grpc_metadata(),
                    )
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.NOT_FOUND:
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise IMPORT_reboot.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: IMPORT_typing.Optional[IMPORT_google_protobuf_any_pb2.Any] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    assert response_or_error is not None
                    assert response_or_error.TypeName() != ""

                    response = rbt.std.collections.v1.sorted_map_pb2.RangeResponse()

                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(IMPORT_uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.std.collections.v1.sorted_map_pb2.RangeResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise IMPORT_asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()


    class RangeAborted(IMPORT_reboot.aio.aborted.Aborted):

        MethodError = IMPORT_typing.Union[
            rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError        ]

        Error = IMPORT_typing.Union[
            MethodError,
            IMPORT_reboot.aio.aborted.GrpcError,
            IMPORT_reboot.aio.aborted.RebootError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
            rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.REBOOT_ERROR_TYPES
        )

        _error: Error
        _code: IMPORT_grpc.StatusCode
        _message: IMPORT_typing.Optional[str]

        def __init__(
            self,
            error: MethodError |  IMPORT_reboot.aio.aborted.GrpcError,
            *,
            message: IMPORT_typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            IMPORT_reboot_aio_types.assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Reboot specific or declared method error.
                code = IMPORT_grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> IMPORT_grpc.StatusCode:
            return self._code

        @property
        def message(self) -> IMPORT_typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: IMPORT_google_rpc_status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: IMPORT_grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: IMPORT_google_protobuf_message.Message) -> bool:
            if message.DESCRIPTOR.full_name == 'rbt.std.collections.v1.InvalidRangeError':
                return True
            return False

    class ReverseRangeTask:
        """Represents a scheduled task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be).
        """

        @classmethod
        def retrieve(
            cls,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ):
            return cls(context, task_id=task_id)

        def __init__(
            self,
            context: IMPORT_reboot_aio_contexts.Context | IMPORT_reboot_aio_external.ExternalContext,
            *,
            task_id: IMPORT_rbt_v1alpha1.tasks_pb2.TaskId,
        ) -> None:
            # Depending on the context type (inside or outside a Reboot application)
            # we may or may not know the application ID. If we don't know it, then
            # the `ExternalContext.gateway` will determine it.
            #
            # TODO: in the future we expect to support cross-application calls, in
            #       which case the developer may explicitly pass in an application ID
            #       here.
            self._application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId] = None
            if isinstance(context, IMPORT_reboot_aio_contexts.Context):
                self._application_id = context.application_id
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> IMPORT_typing.Generator[None, None, rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse]:
            """Awaits for task to finish and returns it's response."""
            async def wait_for_task() -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
                channel = self._channel_manager.get_channel_to_state(
                    self._task_id.state_type,
                    IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                )

                stub = IMPORT_rbt_v1alpha1.tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        IMPORT_rbt_v1alpha1.tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=IMPORT_reboot_aio_headers.Headers(
                            state_ref=IMPORT_reboot_aio_types.StateRef(self._task_id.state_ref),
                            application_id=self._application_id,
                        ).to_grpc_metadata(),
                    )
                except IMPORT_grpc.aio.AioRpcError as error:
                    if error.code() == IMPORT_grpc.StatusCode.NOT_FOUND:
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise IMPORT_reboot.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: IMPORT_typing.Optional[IMPORT_google_protobuf_any_pb2.Any] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    assert response_or_error is not None
                    assert response_or_error.TypeName() != ""

                    response = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse()

                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise IMPORT_reboot.aio.aborted.SystemAborted(
                            IMPORT_rbt_v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(IMPORT_uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise IMPORT_asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()


    class ReverseRangeAborted(IMPORT_reboot.aio.aborted.Aborted):

        MethodError = IMPORT_typing.Union[
            rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError        ]

        Error = IMPORT_typing.Union[
            MethodError,
            IMPORT_reboot.aio.aborted.GrpcError,
            IMPORT_reboot.aio.aborted.RebootError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
            rbt.std.collections.v1.sorted_map_pb2.InvalidRangeError        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES +
            IMPORT_reboot.aio.aborted.REBOOT_ERROR_TYPES
        )

        _error: Error
        _code: IMPORT_grpc.StatusCode
        _message: IMPORT_typing.Optional[str]

        def __init__(
            self,
            error: MethodError |  IMPORT_reboot.aio.aborted.GrpcError,
            *,
            message: IMPORT_typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + IMPORT_reboot.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            IMPORT_reboot_aio_types.assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Reboot specific or declared method error.
                code = IMPORT_grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> IMPORT_grpc.StatusCode:
            return self._code

        @property
        def message(self) -> IMPORT_typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: IMPORT_google_rpc_status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: IMPORT_grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: IMPORT_google_protobuf_message.Message) -> bool:
            if message.DESCRIPTOR.full_name == 'rbt.std.collections.v1.InvalidRangeError':
                return True
            return False


    class WeakReference(IMPORT_typing.Generic[SortedMap_ScheduleTypeVar]):

        _schedule_type: type[SortedMap_ScheduleTypeVar]

        def __init__(
            self,
            # When application ID is None, refers to a state within the application given by the context.
            application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId],
            state_id: IMPORT_reboot_aio_types.StateId,
            *,
            schedule_type: type[SortedMap_ScheduleTypeVar],
            bearer_token: IMPORT_typing.Optional[str] = None,
        ):
            self._application_id = application_id
            self._state_ref = IMPORT_reboot_aio_types.StateRef.from_id(
              SortedMap.__state_type_name__,
              state_id,
            )
            self._schedule_type = schedule_type
            self._idempotency_manager: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.IdempotencyManager] = None
            self._reader_stub: IMPORT_typing.Optional[SortedMapReaderStub] = None
            self._writer_stub: IMPORT_typing.Optional[SortedMapWriterStub] = None
            self._workflow_stub: IMPORT_typing.Optional[SortedMapWorkflowStub] = None
            self._tasks_stub: IMPORT_typing.Optional[SortedMapTasksStub] = None
            self._bearer_token = bearer_token

        @property
        def state_id(self) -> IMPORT_reboot_aio_types.StateId:
            return self._state_ref.id

        def _reader(
            self,
            context: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        ) -> SortedMapReaderStub:
            if self._reader_stub is None:
                self._reader_stub = SortedMapReaderStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._reader_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise IMPORT_reboot_aio_call.MixedContextsError(
                    "This `WeakReference` for `SortedMap` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`SortedMap.ref('{self.state_id}')`."
                )
            return self._reader_stub

        def _writer(
            self,
            context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        ) -> SortedMapWriterStub:
            if self._writer_stub is None:
                self._writer_stub = SortedMapWriterStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._writer_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise IMPORT_reboot_aio_call.MixedContextsError(
                    "This `WeakReference` for `SortedMap` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`SortedMap.ref('{self.state_id}')`."
                )
            return self._writer_stub

        def _workflow(
            self,
            context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        ) -> SortedMapWorkflowStub:
            if self._workflow_stub is None:
                self._workflow_stub = SortedMapWorkflowStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._workflow_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise IMPORT_reboot_aio_call.MixedContextsError(
                    "This `WeakReference` for `SortedMap` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`SortedMap.ref('{self.state_id}')`."
                )
            return self._workflow_stub

        def _tasks(
            self,
            context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        ) -> SortedMapTasksStub:
            if self._tasks_stub is None:
                self._tasks_stub = SortedMapTasksStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._tasks_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise IMPORT_reboot_aio_call.MixedContextsError(
                    "This `WeakReference` for `SortedMap` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`SortedMap.ref('{self.state_id}')`."
                )
            return self._tasks_stub

        class _Reactively:

            def __init__(
                self,
                *,
                application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId],
                state_ref: IMPORT_reboot_aio_types.StateRef,
                bearer_token: IMPORT_typing.Optional[str] = None,
            ):
                self._application_id = application_id
                self._state_ref = state_ref
                self._bearer_token = bearer_token

            async def Get(
                # In methods which are dealing with user input, (i.e.,
                # proto message field names), we should use '__double_underscored__'
                # variables to avoid any potential name conflicts with the method's
                # parameters.
                # The '__self__' parameter is a convention in Python to
                # indicate that this method is a bound method, so we use
                # '__this__' instead.
                __this__,
                # Explicitly-reactive calls only make sense in the context of either...
                # (A) an external client, or...
                # (B) methods that may reasonably run for a long time, which in Reboot means: readers or workflows.
                __context__: IMPORT_reboot_aio_external.ExternalContext | IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WorkflowContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                key: IMPORT_typing.Optional[str] = None,
            ) -> IMPORT_typing.AsyncIterator[rbt.std.collections.v1.sorted_map_pb2.GetResponse]:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_external.ExternalContext, IMPORT_reboot_aio_contexts.ReaderContext, IMPORT_reboot_aio_contexts.WorkflowContext])

                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.GetRequest(
                    key=key,  # type: ignore[arg-type]
                )

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __bearer_token__: IMPORT_typing.Optional[str] = None
                __app_internal_authorization__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                if __bearer_token__ is None:
                    __bearer_token__ = __this__._bearer_token
                if __bearer_token__ is None and isinstance(__context__, IMPORT_reboot_aio_external.ExternalContext):
                    # Within a Reboot context we do not pass on the caller's bearer token, as that might
                    # have security implications - we cannot simply trust any service we are calling with
                    # the user's credentials. Instead, the developer can rely on the default app-internal
                    # auth, or override that and set an explicit bearer token.
                    #
                    # In the case of `ExternalContext`, however, its `bearer_token` was set specifically
                    # by the developer for the purpose of making these calls. Note that only
                    # `ExternalContext` even has a `bearer_token` field.
                    __bearer_token__ = __context__.bearer_token

                if __metadata__ is None:
                    __metadata__ = ()

                if isinstance(__context__, IMPORT_reboot_aio_contexts.Context):
                    if __this__._application_id is None:
                        # Given our context type (inside a Reboot application) we can default to
                        # making the application send traffic to itself.
                        __this__._application_id = __context__.application_id
                        # It is safe to use app-internal auth given we know we're talking to
                        # the same application.
                        __app_internal_authorization__ = __context__._app_internal_api_key_secret

                __headers__ = IMPORT_reboot_aio_headers.Headers(
                    bearer_token=__bearer_token__,
                    state_ref=__this__._state_ref,
                    application_id=__this__._application_id,
                    app_internal_authorization=__app_internal_authorization__
                )

                __metadata__ += __headers__.to_grpc_metadata()

                __query_backoff__ = IMPORT_reboot_aio_backoff.Backoff()
                while True:
                    try:
                        async with __context__.channel_manager.get_channel_to_state(
                            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            __this__._state_ref,
                        ) as __channel__:

                            __call__ = IMPORT_rbt_v1alpha1.react_pb2_grpc.ReactStub(
                                __channel__
                            ).Query(
                                IMPORT_rbt_v1alpha1.react_pb2.QueryRequest(
                                    method='Get',
                                    request=__request__.SerializeToString(),
                                ),
                                metadata=__metadata__,
                            )

                            async for __query_response__ in __call__:
                                # Clear the backoff so we don't wait
                                # as long the next time we get
                                # disconnected.
                                __query_backoff__.clear()

                                # The backend may have sent us this query
                                # response only to let us know that a new
                                # idempotency key has been recorded; there may
                                # not be a new response. Python callers don't
                                # (currently) care about such an event, so we
                                # simply ignore it.
                                if not __query_response__.HasField("response"):
                                    continue

                                __response__ = rbt.std.collections.v1.sorted_map_pb2.GetResponse()
                                __response__.ParseFromString(__query_response__.response)
                                yield __response__

                    except BaseException as exception:
                        # We expect to get disconnected from the placement
                        # planner from time to time, e.g., when it is
                        # being updated, but we don't want that error to
                        # propagate, we just want to retry.
                        if IMPORT_reboot.aio.aborted.is_grpc_retryable_exception(exception):
                            await __query_backoff__()
                            continue
                        raise

            async def Range(
                # In methods which are dealing with user input, (i.e.,
                # proto message field names), we should use '__double_underscored__'
                # variables to avoid any potential name conflicts with the method's
                # parameters.
                # The '__self__' parameter is a convention in Python to
                # indicate that this method is a bound method, so we use
                # '__this__' instead.
                __this__,
                # Explicitly-reactive calls only make sense in the context of either...
                # (A) an external client, or...
                # (B) methods that may reasonably run for a long time, which in Reboot means: readers or workflows.
                __context__: IMPORT_reboot_aio_external.ExternalContext | IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WorkflowContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_typing.AsyncIterator[rbt.std.collections.v1.sorted_map_pb2.RangeResponse]:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_external.ExternalContext, IMPORT_reboot_aio_contexts.ReaderContext, IMPORT_reboot_aio_contexts.WorkflowContext])

                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __bearer_token__: IMPORT_typing.Optional[str] = None
                __app_internal_authorization__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                if __bearer_token__ is None:
                    __bearer_token__ = __this__._bearer_token
                if __bearer_token__ is None and isinstance(__context__, IMPORT_reboot_aio_external.ExternalContext):
                    # Within a Reboot context we do not pass on the caller's bearer token, as that might
                    # have security implications - we cannot simply trust any service we are calling with
                    # the user's credentials. Instead, the developer can rely on the default app-internal
                    # auth, or override that and set an explicit bearer token.
                    #
                    # In the case of `ExternalContext`, however, its `bearer_token` was set specifically
                    # by the developer for the purpose of making these calls. Note that only
                    # `ExternalContext` even has a `bearer_token` field.
                    __bearer_token__ = __context__.bearer_token

                if __metadata__ is None:
                    __metadata__ = ()

                if isinstance(__context__, IMPORT_reboot_aio_contexts.Context):
                    if __this__._application_id is None:
                        # Given our context type (inside a Reboot application) we can default to
                        # making the application send traffic to itself.
                        __this__._application_id = __context__.application_id
                        # It is safe to use app-internal auth given we know we're talking to
                        # the same application.
                        __app_internal_authorization__ = __context__._app_internal_api_key_secret

                __headers__ = IMPORT_reboot_aio_headers.Headers(
                    bearer_token=__bearer_token__,
                    state_ref=__this__._state_ref,
                    application_id=__this__._application_id,
                    app_internal_authorization=__app_internal_authorization__
                )

                __metadata__ += __headers__.to_grpc_metadata()

                __query_backoff__ = IMPORT_reboot_aio_backoff.Backoff()
                while True:
                    try:
                        async with __context__.channel_manager.get_channel_to_state(
                            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            __this__._state_ref,
                        ) as __channel__:

                            __call__ = IMPORT_rbt_v1alpha1.react_pb2_grpc.ReactStub(
                                __channel__
                            ).Query(
                                IMPORT_rbt_v1alpha1.react_pb2.QueryRequest(
                                    method='Range',
                                    request=__request__.SerializeToString(),
                                ),
                                metadata=__metadata__,
                            )

                            async for __query_response__ in __call__:
                                # Clear the backoff so we don't wait
                                # as long the next time we get
                                # disconnected.
                                __query_backoff__.clear()

                                # The backend may have sent us this query
                                # response only to let us know that a new
                                # idempotency key has been recorded; there may
                                # not be a new response. Python callers don't
                                # (currently) care about such an event, so we
                                # simply ignore it.
                                if not __query_response__.HasField("response"):
                                    continue

                                __response__ = rbt.std.collections.v1.sorted_map_pb2.RangeResponse()
                                __response__.ParseFromString(__query_response__.response)
                                yield __response__

                    except BaseException as exception:
                        # We expect to get disconnected from the placement
                        # planner from time to time, e.g., when it is
                        # being updated, but we don't want that error to
                        # propagate, we just want to retry.
                        if IMPORT_reboot.aio.aborted.is_grpc_retryable_exception(exception):
                            await __query_backoff__()
                            continue
                        raise

            async def ReverseRange(
                # In methods which are dealing with user input, (i.e.,
                # proto message field names), we should use '__double_underscored__'
                # variables to avoid any potential name conflicts with the method's
                # parameters.
                # The '__self__' parameter is a convention in Python to
                # indicate that this method is a bound method, so we use
                # '__this__' instead.
                __this__,
                # Explicitly-reactive calls only make sense in the context of either...
                # (A) an external client, or...
                # (B) methods that may reasonably run for a long time, which in Reboot means: readers or workflows.
                __context__: IMPORT_reboot_aio_external.ExternalContext | IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WorkflowContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_typing.AsyncIterator[rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse]:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_external.ExternalContext, IMPORT_reboot_aio_contexts.ReaderContext, IMPORT_reboot_aio_contexts.WorkflowContext])

                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __bearer_token__: IMPORT_typing.Optional[str] = None
                __app_internal_authorization__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                if __bearer_token__ is None:
                    __bearer_token__ = __this__._bearer_token
                if __bearer_token__ is None and isinstance(__context__, IMPORT_reboot_aio_external.ExternalContext):
                    # Within a Reboot context we do not pass on the caller's bearer token, as that might
                    # have security implications - we cannot simply trust any service we are calling with
                    # the user's credentials. Instead, the developer can rely on the default app-internal
                    # auth, or override that and set an explicit bearer token.
                    #
                    # In the case of `ExternalContext`, however, its `bearer_token` was set specifically
                    # by the developer for the purpose of making these calls. Note that only
                    # `ExternalContext` even has a `bearer_token` field.
                    __bearer_token__ = __context__.bearer_token

                if __metadata__ is None:
                    __metadata__ = ()

                if isinstance(__context__, IMPORT_reboot_aio_contexts.Context):
                    if __this__._application_id is None:
                        # Given our context type (inside a Reboot application) we can default to
                        # making the application send traffic to itself.
                        __this__._application_id = __context__.application_id
                        # It is safe to use app-internal auth given we know we're talking to
                        # the same application.
                        __app_internal_authorization__ = __context__._app_internal_api_key_secret

                __headers__ = IMPORT_reboot_aio_headers.Headers(
                    bearer_token=__bearer_token__,
                    state_ref=__this__._state_ref,
                    application_id=__this__._application_id,
                    app_internal_authorization=__app_internal_authorization__
                )

                __metadata__ += __headers__.to_grpc_metadata()

                __query_backoff__ = IMPORT_reboot_aio_backoff.Backoff()
                while True:
                    try:
                        async with __context__.channel_manager.get_channel_to_state(
                            IMPORT_reboot_aio_types.StateTypeName('rbt.std.collections.v1.SortedMap'),
                            __this__._state_ref,
                        ) as __channel__:

                            __call__ = IMPORT_rbt_v1alpha1.react_pb2_grpc.ReactStub(
                                __channel__
                            ).Query(
                                IMPORT_rbt_v1alpha1.react_pb2.QueryRequest(
                                    method='ReverseRange',
                                    request=__request__.SerializeToString(),
                                ),
                                metadata=__metadata__,
                            )

                            async for __query_response__ in __call__:
                                # Clear the backoff so we don't wait
                                # as long the next time we get
                                # disconnected.
                                __query_backoff__.clear()

                                # The backend may have sent us this query
                                # response only to let us know that a new
                                # idempotency key has been recorded; there may
                                # not be a new response. Python callers don't
                                # (currently) care about such an event, so we
                                # simply ignore it.
                                if not __query_response__.HasField("response"):
                                    continue

                                __response__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse()
                                __response__.ParseFromString(__query_response__.response)
                                yield __response__

                    except BaseException as exception:
                        # We expect to get disconnected from the placement
                        # planner from time to time, e.g., when it is
                        # being updated, but we don't want that error to
                        # propagate, we just want to retry.
                        if IMPORT_reboot.aio.aborted.is_grpc_retryable_exception(exception):
                            await __query_backoff__()
                            continue
                        raise


        def reactively(self):
            return SortedMap.WeakReference._Reactively(
                application_id=self._application_id,
                state_ref=self._state_ref,
                bearer_token=self._bearer_token,
            )

        class _Idempotently(IMPORT_typing.Generic[SortedMap_IdempotentlyScheduleTypeVar]):

            _weak_reference: SortedMap.WeakReference[SortedMap_IdempotentlyScheduleTypeVar]

            def __init__(
                self,
                *,
                weak_reference: SortedMap.WeakReference[SortedMap_IdempotentlyScheduleTypeVar],
                idempotency: IMPORT_reboot_aio_idempotency.Idempotency,
            ):
                self._weak_reference = weak_reference
                self._idempotency = idempotency

            def schedule(
                self,
                *,
                when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
            ) -> SortedMap_IdempotentlyScheduleTypeVar:
                return self._weak_reference._schedule_type(
                    self._weak_reference._application_id,
                    self._weak_reference._tasks,
                    when=when,
                    idempotency=self._idempotency,
                )

            def spawn(
                self,
                *,
                when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
            ) -> SortedMap.WeakReference._Spawn:
                return SortedMap.WeakReference._Spawn(
                    self._weak_reference._application_id,
                    self._weak_reference._tasks,
                    when=when,
                    idempotency=self._idempotency,
                )

            async def Insert(
                # In methods which are dealing with user input, (i.e.,
                # proto message field names), we should use '__double_underscored__'
                # variables to avoid any potential name conflicts with the method's
                # parameters.
                # The '__self__' parameter is a convention in Python to
                # indicate that this method is a bound method, so we use
                # '__this__' instead.
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                entries: IMPORT_typing.Optional[dict[str, bytes]] = None,
            ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)

                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.idempotency is not None:
                    raise RuntimeError(
                        'Found redundant idempotency in `Options`'
                    )

                __options__ = IMPORT_dataclasses.replace(
                    __options__,
                    idempotency_key=__this__._idempotency.key,
                    idempotency_alias=__this__._idempotency.alias,
                    generate_idempotency=__this__._idempotency.generate,
                )

                return await __this__._weak_reference.Insert(
                    __context__,
                    __options__,
                    entries=entries,
                )

            async def Remove(
                # In methods which are dealing with user input, (i.e.,
                # proto message field names), we should use '__double_underscored__'
                # variables to avoid any potential name conflicts with the method's
                # parameters.
                # The '__self__' parameter is a convention in Python to
                # indicate that this method is a bound method, so we use
                # '__this__' instead.
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_reboot_aio_call.Options = IMPORT_reboot_aio_call.Options(),
                *,
                keys: IMPORT_typing.Optional[IMPORT_typing.Iterable[str]] = None,
            ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)

                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.idempotency is not None:
                    raise RuntimeError(
                        'Found redundant idempotency in `Options`'
                    )

                __options__ = IMPORT_dataclasses.replace(
                    __options__,
                    idempotency_key=__this__._idempotency.key,
                    idempotency_alias=__this__._idempotency.alias,
                    generate_idempotency=__this__._idempotency.generate,
                )

                return await __this__._weak_reference.Remove(
                    __context__,
                    __options__,
                    keys=keys,
                )


        def idempotently(
            self,
            alias: IMPORT_typing.Optional[str] = None,
            *,
            key: IMPORT_typing.Optional[IMPORT_uuid.UUID] = None,
            each_iteration: IMPORT_typing.Optional[bool] = None,
        ) -> SortedMap.WeakReference._Idempotently[SortedMap_ScheduleTypeVar]:
            return SortedMap.WeakReference._Idempotently(
                weak_reference=self,
                idempotency=IMPORT_reboot_aio_contexts.Context.idempotency(
                    key=key, alias=alias, each_iteration=each_iteration
                )
            )

        def unidempotently(self):
            return self.idempotently(key=IMPORT_uuid.uuid4())

        def schedule(
            self,
            *,
            when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
        ) -> SortedMap_ScheduleTypeVar:
            return self._schedule_type(self._application_id, self._tasks, when=when)

        class _Schedule:

            def __init__(
                self,
                application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId],
                tasks: IMPORT_typing.Callable[[IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext], SortedMapTasksStub],
                *,
                when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
                idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
            ) -> None:
                self._application_id = application_id
                self._tasks = tasks
                self._when = ensure_has_timezone(when=when)
                self._idempotency = idempotency

            # SortedMap callable tasks:
            async def Insert(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                entries: IMPORT_typing.Optional[dict[str, bytes]] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.TransactionContext])

                if entries is not None and not isinstance(
                    entries,
                    dict,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.InsertRequest': field 'entries' is not "
                        f"of required type 'dict[str, bytes]'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.InsertRequest(
                    entries=entries,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Insert(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return __task_id__

            async def Remove(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                keys: IMPORT_typing.Optional[IMPORT_typing.Iterable[str]] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.TransactionContext])

                if keys is not None and not isinstance(
                    keys,
                    IMPORT_typing.Iterable,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RemoveRequest': field 'keys' is not "
                        f"of required type 'typing.Iterable[str]'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RemoveRequest(
                    keys=keys,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Remove(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return __task_id__

            async def Get(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                key: IMPORT_typing.Optional[str] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.TransactionContext])

                if key is not None and not isinstance(
                    key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.GetRequest': field 'key' is not "
                        f"of required type 'str'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.GetRequest(
                    key=key,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Get(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return __task_id__

            async def Range(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.TransactionContext])

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Range(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return __task_id__

            async def ReverseRange(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.TransactionContext])

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).ReverseRange(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return __task_id__


        # A `WriterContext` can not call any methods in `_Schedule` to
        # prevent a writer from doing a `Foo.ref()` and trying to
        # schedule. However, we want to allow a writer to schedule
        # when we are constructing a `WeakReference` from
        # `self.ref()` so instead we return a `_WriterSchedule` to
        # provide type safety that allows a `WriterContext` to
        # schedule (for itself).
        class _WriterSchedule:

            def __init__(
                self,
                application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId],
                tasks: IMPORT_typing.Callable[[IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext], SortedMapTasksStub],
                *,
                when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
                idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
            ) -> None:
                self._tasks = tasks
                self._when = ensure_has_timezone(when=when)
                self._idempotency = idempotency

            # SortedMap callable tasks:
            async def Insert(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                entries: IMPORT_typing.Optional[dict[str, bytes]] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # Only `writer`s and `transaction`s should ``schedule()`, a
                # `workflow` should `spawn()`.
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext])

                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)

                if entries is not None and not isinstance(
                    entries,
                    dict,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.InsertRequest': field 'entries' is not "
                        f"of required type 'dict[str, bytes]'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.InsertRequest(
                    entries=entries,  # type: ignore[arg-type]
                )
                if isinstance(__context__, IMPORT_reboot_aio_contexts.WriterContext):
                    return SortedMapServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Insert(
                        __request__,
                        schedule=__this__._when,
                    ).task_id

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency is not None:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                return await __this__._tasks(
                    __context__
                ).Insert(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

            async def Remove(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                keys: IMPORT_typing.Optional[IMPORT_typing.Iterable[str]] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # Only `writer`s and `transaction`s should ``schedule()`, a
                # `workflow` should `spawn()`.
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext])

                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)

                if keys is not None and not isinstance(
                    keys,
                    IMPORT_typing.Iterable,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RemoveRequest': field 'keys' is not "
                        f"of required type 'typing.Iterable[str]'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RemoveRequest(
                    keys=keys,  # type: ignore[arg-type]
                )
                if isinstance(__context__, IMPORT_reboot_aio_contexts.WriterContext):
                    return SortedMapServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Remove(
                        __request__,
                        schedule=__this__._when,
                    ).task_id

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency is not None:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                return await __this__._tasks(
                    __context__
                ).Remove(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

            async def Get(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                key: IMPORT_typing.Optional[str] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # Only `writer`s and `transaction`s should ``schedule()`, a
                # `workflow` should `spawn()`.
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext])

                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)

                if key is not None and not isinstance(
                    key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.GetRequest': field 'key' is not "
                        f"of required type 'str'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.GetRequest(
                    key=key,  # type: ignore[arg-type]
                )
                if isinstance(__context__, IMPORT_reboot_aio_contexts.WriterContext):
                    return SortedMapServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Get(
                        __request__,
                        schedule=__this__._when,
                    ).task_id

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency is not None:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                return await __this__._tasks(
                    __context__
                ).Get(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

            async def Range(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # Only `writer`s and `transaction`s should ``schedule()`, a
                # `workflow` should `spawn()`.
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext])

                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )
                if isinstance(__context__, IMPORT_reboot_aio_contexts.WriterContext):
                    return SortedMapServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Range(
                        __request__,
                        schedule=__this__._when,
                    ).task_id

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency is not None:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                return await __this__._tasks(
                    __context__
                ).Range(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

            async def ReverseRange(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> IMPORT_rbt_v1alpha1.tasks_pb2.TaskId:
                # Only `writer`s and `transaction`s should ``schedule()`, a
                # `workflow` should `spawn()`.
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WriterContext, IMPORT_reboot_aio_contexts.TransactionContext])

                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )
                if isinstance(__context__, IMPORT_reboot_aio_contexts.WriterContext):
                    return SortedMapServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).ReverseRange(
                        __request__,
                        schedule=__this__._when,
                    ).task_id

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency is not None:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                return await __this__._tasks(
                    __context__
                ).ReverseRange(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )


        def spawn(
            self,
            *,
            when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
        ) -> SortedMap.WeakReference._Spawn:
            return SortedMap.WeakReference._Spawn(
                self._application_id, self._tasks, when=when
            )

        class _Spawn:

            def __init__(
                self,
                application_id: IMPORT_typing.Optional[IMPORT_reboot_aio_types.ApplicationId],
                tasks: IMPORT_typing.Callable[[IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext], SortedMapTasksStub],
                *,
                when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
                idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None,
            ) -> None:
                self._application_id = application_id
                self._tasks = tasks
                self._when = ensure_has_timezone(when=when)
                self._idempotency = idempotency

            # SortedMap callable tasks:
            async def Insert(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                entries: IMPORT_typing.Optional[dict[str, bytes]] = None,
            ) -> SortedMap.InsertTask:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])

                if entries is not None and not isinstance(
                    entries,
                    dict,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.InsertRequest': field 'entries' is not "
                        f"of required type 'dict[str, bytes]'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.InsertRequest(
                    entries=entries,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Insert(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return SortedMap.InsertTask(
                    __context__,
                    task_id=__task_id__,
                )

            async def Remove(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                keys: IMPORT_typing.Optional[IMPORT_typing.Iterable[str]] = None,
            ) -> SortedMap.RemoveTask:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])

                if keys is not None and not isinstance(
                    keys,
                    IMPORT_typing.Iterable,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RemoveRequest': field 'keys' is not "
                        f"of required type 'typing.Iterable[str]'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RemoveRequest(
                    keys=keys,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Remove(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return SortedMap.RemoveTask(
                    __context__,
                    task_id=__task_id__,
                )

            async def Get(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                key: IMPORT_typing.Optional[str] = None,
            ) -> SortedMap.GetTask:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])

                if key is not None and not isinstance(
                    key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.GetRequest': field 'key' is not "
                        f"of required type 'str'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.GetRequest(
                    key=key,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Get(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return SortedMap.GetTask(
                    __context__,
                    task_id=__task_id__,
                )

            async def Range(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> SortedMap.RangeTask:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.RangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).Range(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return SortedMap.RangeTask(
                    __context__,
                    task_id=__task_id__,
                )

            async def ReverseRange(
                __this__,
                __context__: IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
                __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
                *,
                start_key: IMPORT_typing.Optional[str] = None,
                end_key: IMPORT_typing.Optional[str] = None,
                limit: IMPORT_typing.Optional[int] = None,
            ) -> SortedMap.ReverseRangeTask:
                # UX improvement: check that neither positional argument was accidentally
                # given a gRPC request type.
                IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
                IMPORT_reboot_aio_types.assert_type(__context__, [IMPORT_reboot_aio_contexts.WorkflowContext, IMPORT_reboot_aio_external.ExternalContext])

                if start_key is not None and not isinstance(
                    start_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'start_key' is not "
                        f"of required type 'str'"
                    )
                if end_key is not None and not isinstance(
                    end_key,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'end_key' is not "
                        f"of required type 'str'"
                    )
                if limit is not None and not isinstance(
                    limit,
                    int,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'limit' is not "
                        f"of required type 'int'"
                    )
                # TODO: mypy-protobuf declares that
                # `IMPORT_google_protobuf_message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                __request__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest(
                    start_key=start_key,  # type: ignore[arg-type]
                    end_key=end_key,  # type: ignore[arg-type]
                    limit=limit,  # type: ignore[arg-type]
                )

                __schedule__: IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone] = (IMPORT_reboot_time_DateTimeWithTimeZone.now() + __this__._when) if isinstance(
                    __this__._when, IMPORT_datetime_timedelta
                ) else __this__._when

                __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
                __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = __this__._idempotency
                __bearer_token__: IMPORT_typing.Optional[str] = None

                if __options__ is not None:
                    IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                    if __options__.idempotency:
                        if __idempotency__ is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        __idempotency__ = __options__.idempotency
                    if __options__.metadata is not None:
                        __metadata__ = __options__.metadata
                    if __options__.bearer_token is not None:
                        __bearer_token__ = __options__.bearer_token

                # Add scheduling information to the metadata.
                __metadata__ = (
                    (IMPORT_reboot_aio_headers.TASK_SCHEDULE,
                    __schedule__.isoformat() if __schedule__ else ''),
                ) + (__metadata__ or tuple())

                __task_id__ = await __this__._tasks(
                    __context__
                ).ReverseRange(
                    __request__,
                    idempotency=__idempotency__,
                    metadata=__metadata__,
                    bearer_token=__bearer_token__,
                )

                return SortedMap.ReverseRangeTask(
                    __context__,
                    task_id=__task_id__,
                )


        # SortedMap specific methods:
        async def Insert(
            __this__,
            __context__: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
            __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
            *,
            entries: IMPORT_typing.Optional[dict[str, bytes]] = None,
        ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
            # UX improvement: check that neither positional argument was accidentally
            # given a gRPC request type.
            IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)
            IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.InsertRequest)

            if entries is not None and not isinstance(
                entries,
                dict,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.InsertRequest': field 'entries' is not "
                    f"of required type 'dict[str, bytes]'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            __request__ = rbt.std.collections.v1.sorted_map_pb2.InsertRequest(
                entries=entries,  # type: ignore[arg-type]
            )
            __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None
            __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
            __bearer_token__: IMPORT_typing.Optional[str] = None
            if __options__ is not None:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.idempotency is not None:
                    __idempotency__ = __options__.idempotency
                if __options__.metadata is not None:
                    __metadata__ = __options__.metadata
                if __options__.bearer_token is not None:
                    __bearer_token__ = __options__.bearer_token

            return await __this__._writer(__context__).Insert(
                __request__,
                idempotency=__idempotency__,
                metadata=__metadata__,
                bearer_token=__bearer_token__,
            )

        async def Remove(
            __this__,
            __context__: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
            __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
            *,
            keys: IMPORT_typing.Optional[IMPORT_typing.Iterable[str]] = None,
        ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
            # UX improvement: check that neither positional argument was accidentally
            # given a gRPC request type.
            IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)
            IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RemoveRequest)

            if keys is not None and not isinstance(
                keys,
                IMPORT_typing.Iterable,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.RemoveRequest': field 'keys' is not "
                    f"of required type 'typing.Iterable[str]'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            __request__ = rbt.std.collections.v1.sorted_map_pb2.RemoveRequest(
                keys=keys,  # type: ignore[arg-type]
            )
            __idempotency__: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None
            __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
            __bearer_token__: IMPORT_typing.Optional[str] = None
            if __options__ is not None:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.idempotency is not None:
                    __idempotency__ = __options__.idempotency
                if __options__.metadata is not None:
                    __metadata__ = __options__.metadata
                if __options__.bearer_token is not None:
                    __bearer_token__ = __options__.bearer_token

            return await __this__._writer(__context__).Remove(
                __request__,
                idempotency=__idempotency__,
                metadata=__metadata__,
                bearer_token=__bearer_token__,
            )

        async def Get(
            __this__,
            __context__: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
            __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
            *,
            key: IMPORT_typing.Optional[str] = None,
        ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
            # UX improvement: check that neither positional argument was accidentally
            # given a gRPC request type.
            IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)
            IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.GetRequest)

            if key is not None and not isinstance(
                key,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.GetRequest': field 'key' is not "
                    f"of required type 'str'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            __request__ = rbt.std.collections.v1.sorted_map_pb2.GetRequest(
                key=key,  # type: ignore[arg-type]
            )
            __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
            __bearer_token__: IMPORT_typing.Optional[str] = None
            if __options__ is not None:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.metadata is not None:
                    __metadata__ = __options__.metadata
                if __options__.bearer_token is not None:
                    __bearer_token__ = __options__.bearer_token
            return await __this__._reader(__context__).Get(
                __request__,
                metadata=__metadata__,
                bearer_token=__bearer_token__,
            )
        async def Range(
            __this__,
            __context__: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
            __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
            *,
            start_key: IMPORT_typing.Optional[str] = None,
            end_key: IMPORT_typing.Optional[str] = None,
            limit: IMPORT_typing.Optional[int] = None,
        ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
            # UX improvement: check that neither positional argument was accidentally
            # given a gRPC request type.
            IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)
            IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.RangeRequest)

            if start_key is not None and not isinstance(
                start_key,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'start_key' is not "
                    f"of required type 'str'"
                )
            if end_key is not None and not isinstance(
                end_key,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'end_key' is not "
                    f"of required type 'str'"
                )
            if limit is not None and not isinstance(
                limit,
                int,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.RangeRequest': field 'limit' is not "
                    f"of required type 'int'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            __request__ = rbt.std.collections.v1.sorted_map_pb2.RangeRequest(
                start_key=start_key,  # type: ignore[arg-type]
                end_key=end_key,  # type: ignore[arg-type]
                limit=limit,  # type: ignore[arg-type]
            )
            __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
            __bearer_token__: IMPORT_typing.Optional[str] = None
            if __options__ is not None:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.metadata is not None:
                    __metadata__ = __options__.metadata
                if __options__.bearer_token is not None:
                    __bearer_token__ = __options__.bearer_token
            return await __this__._reader(__context__).Range(
                __request__,
                metadata=__metadata__,
                bearer_token=__bearer_token__,
            )
        async def ReverseRange(
            __this__,
            __context__: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
            __options__: IMPORT_typing.Optional[IMPORT_reboot_aio_call.Options] = None,
            *,
            start_key: IMPORT_typing.Optional[str] = None,
            end_key: IMPORT_typing.Optional[str] = None,
            limit: IMPORT_typing.Optional[int] = None,
        ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
            # UX improvement: check that neither positional argument was accidentally
            # given a gRPC request type.
            IMPORT_reboot_aio_types.assert_not_request_type(__context__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)
            IMPORT_reboot_aio_types.assert_not_request_type(__options__, request_type=rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest)

            if start_key is not None and not isinstance(
                start_key,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'start_key' is not "
                    f"of required type 'str'"
                )
            if end_key is not None and not isinstance(
                end_key,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'end_key' is not "
                    f"of required type 'str'"
                )
            if limit is not None and not isinstance(
                limit,
                int,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest': field 'limit' is not "
                    f"of required type 'int'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            __request__ = rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest(
                start_key=start_key,  # type: ignore[arg-type]
                end_key=end_key,  # type: ignore[arg-type]
                limit=limit,  # type: ignore[arg-type]
            )
            __metadata__: IMPORT_typing.Optional[IMPORT_reboot_aio_types.GrpcMetadata] = None
            __bearer_token__: IMPORT_typing.Optional[str] = None
            if __options__ is not None:
                IMPORT_reboot_aio_types.assert_type(__options__, [IMPORT_reboot_aio_call.Options])
                if __options__.metadata is not None:
                    __metadata__ = __options__.metadata
                if __options__.bearer_token is not None:
                    __bearer_token__ = __options__.bearer_token
            return await __this__._reader(__context__).ReverseRange(
                __request__,
                metadata=__metadata__,
                bearer_token=__bearer_token__,
            )

    @classmethod
    def ref(
        cls,
        state_id: IMPORT_reboot_aio_types.StateId,
        *,
        bearer_token: IMPORT_typing.Optional[str] = None,
    ) -> SortedMap.WeakReference[SortedMap.WeakReference._Schedule]:
        return SortedMap.WeakReference(
            # TODO(https://github.com/reboot-dev/mono/issues/3226): add support for calling other applications.
            # For now this always stays within the application that creates the context.
            application_id=None,
            state_id=state_id,
            schedule_type=SortedMap.WeakReference._Schedule,
            bearer_token=bearer_token,
        )


    @IMPORT_typing.overload
    @classmethod
    def idempotently(cls, alias: IMPORT_typing.Optional[str] = None, *, each_iteration: bool = False) -> SortedMap._ConstructIdempotently:
        ...

    @IMPORT_typing.overload
    @classmethod
    def idempotently(cls, *, key: IMPORT_uuid.UUID) -> SortedMap._ConstructIdempotently:
        ...

    @classmethod
    def idempotently(
        cls,
        alias: IMPORT_typing.Optional[str] = None,
        *,
        key: IMPORT_typing.Optional[IMPORT_uuid.UUID] = None,
        each_iteration: IMPORT_typing.Optional[bool] = None,
    ) -> SortedMap._ConstructIdempotently:
        return SortedMap._ConstructIdempotently(
            _idempotency=IMPORT_reboot_aio_contexts.Context.idempotency(
                alias=alias, key=key, each_iteration=each_iteration
            ),
        )

    @IMPORT_dataclasses.dataclass(frozen=True)
    class _ConstructIdempotently:

        _idempotency: IMPORT_reboot_aio_idempotency.Idempotency



############################ Servicer Node adapters ############################
# Used by Node.js servicer implementations to access Python code and vice-versa.
# Relevant to servicers, irrelevant to clients.

class SortedMapServicerNodeAdaptor(SortedMap.Servicer):

    def __init__(self):
        self._js_servicer_reference = self._construct_js_servicer()  # type: ignore[attr-defined]

    def authorizer(self) -> IMPORT_typing.Optional[IMPORT_reboot.aio.auth.authorizers.Authorizer]:
        return self._construct_authorizer(self._js_servicer_reference)  # type: ignore[attr-defined]

    async def _read(self, context: IMPORT_reboot_aio_contexts.WorkflowContext) -> str:
        return IMPORT_google_protobuf_json_format.MessageToJson(
            await super().state.read(context)
        )

    async def _write(
        self,
        context: IMPORT_reboot_aio_contexts.WorkflowContext,
        writer: IMPORT_typing.Callable[[str], IMPORT_typing.Awaitable[str]],
        json_options: str,
    ) -> str:

        async def _writer(state: IMPORT_google_protobuf_message.Message):
            json_result_state = await writer(
                IMPORT_google_protobuf_json_format.MessageToJson(state)
            )

            result_state = IMPORT_json.loads(json_result_state)

            state.CopyFrom(
                IMPORT_google_protobuf_json_format.ParseDict(
                    result_state['state'],
                    self.__state_type__(),
                )
            )

            assert 'result' in result_state
            result = result_state['result']
            assert type(result) == str
            return result

        options = IMPORT_json.loads(json_options)

        assert 'idempotency' in options

        return await super().state.idempotently(
            alias=options['idempotency'].get('alias'),
            key=options['idempotency'].get('key'),
            each_iteration=options['idempotency'].get('eachIteration', None),
        )._write(
            context,
            _writer,
            type_result=str,
            check_type=True,
            unidempotently=options.get('unidempotently', False),
        )

    # SortedMap specific methods:
    async def Insert(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.InsertRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.InsertResponse:
        json_state = IMPORT_google_protobuf_json_format.MessageToJson(state)
        json_request = IMPORT_google_protobuf_json_format.MessageToJson(request)

        aborted: IMPORT_asyncio.Future[None] = IMPORT_asyncio.Future()

        try:
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'writer',
                'Insert',
                context,
                aborted,
                json_state,
                json_request,
            )
        except IMPORT_asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = IMPORT_json.loads(json_result)

            if 'effects' in result:
                effects = result['effects']

                state.CopyFrom(
                    IMPORT_google_protobuf_json_format.ParseDict(
                        effects['state'],
                        rbt.std.collections.v1.sorted_map_pb2.SortedMap(),
                    )
                )

                assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

                return IMPORT_google_protobuf_json_format.ParseDict(
                    effects['response'],
                    rbt.std.collections.v1.sorted_map_pb2.InsertResponse(),
                )
            elif 'status' in result:
                raise (
                    SortedMap
                    .InsertAborted
                    .from_status(
                        IMPORT_google_protobuf_json_format.ParseDict(
                            result['status'],
                            IMPORT_google_rpc_status_pb2.Status(),
                        )
                    )
                )
        raise RuntimeError("Unexpected result from Insert")

    async def Remove(
        self,
        context: IMPORT_reboot_aio_contexts.WriterContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RemoveRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RemoveResponse:
        json_state = IMPORT_google_protobuf_json_format.MessageToJson(state)
        json_request = IMPORT_google_protobuf_json_format.MessageToJson(request)

        aborted: IMPORT_asyncio.Future[None] = IMPORT_asyncio.Future()

        try:
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'writer',
                'Remove',
                context,
                aborted,
                json_state,
                json_request,
            )
        except IMPORT_asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = IMPORT_json.loads(json_result)

            if 'effects' in result:
                effects = result['effects']

                state.CopyFrom(
                    IMPORT_google_protobuf_json_format.ParseDict(
                        effects['state'],
                        rbt.std.collections.v1.sorted_map_pb2.SortedMap(),
                    )
                )

                assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

                return IMPORT_google_protobuf_json_format.ParseDict(
                    effects['response'],
                    rbt.std.collections.v1.sorted_map_pb2.RemoveResponse(),
                )
            elif 'status' in result:
                raise (
                    SortedMap
                    .RemoveAborted
                    .from_status(
                        IMPORT_google_protobuf_json_format.ParseDict(
                            result['status'],
                            IMPORT_google_rpc_status_pb2.Status(),
                        )
                    )
                )
        raise RuntimeError("Unexpected result from Remove")

    async def Get(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.GetRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.GetResponse:
        json_state = IMPORT_google_protobuf_json_format.MessageToJson(state)
        json_request = IMPORT_google_protobuf_json_format.MessageToJson(request)

        aborted: IMPORT_asyncio.Future[None] = IMPORT_asyncio.Future()

        try:
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'reader',
                'Get',
                context,
                aborted,
                json_state,
                json_request,
            )
        except IMPORT_asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = IMPORT_json.loads(json_result)

            if 'response' in result:
                return IMPORT_google_protobuf_json_format.ParseDict(
                    result['response'],
                    rbt.std.collections.v1.sorted_map_pb2.GetResponse(),
                )
            elif 'status' in result:
                raise (
                    SortedMap
                    .GetAborted
                    .from_status(
                        IMPORT_google_protobuf_json_format.ParseDict(
                            result['status'],
                            IMPORT_google_rpc_status_pb2.Status(),
                        )
                    )
                )
        raise RuntimeError("Unexpected result from Get")

    async def Range(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.RangeRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.RangeResponse:
        json_state = IMPORT_google_protobuf_json_format.MessageToJson(state)
        json_request = IMPORT_google_protobuf_json_format.MessageToJson(request)

        aborted: IMPORT_asyncio.Future[None] = IMPORT_asyncio.Future()

        try:
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'reader',
                'Range',
                context,
                aborted,
                json_state,
                json_request,
            )
        except IMPORT_asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = IMPORT_json.loads(json_result)

            if 'response' in result:
                return IMPORT_google_protobuf_json_format.ParseDict(
                    result['response'],
                    rbt.std.collections.v1.sorted_map_pb2.RangeResponse(),
                )
            elif 'status' in result:
                raise (
                    SortedMap
                    .RangeAborted
                    .from_status(
                        IMPORT_google_protobuf_json_format.ParseDict(
                            result['status'],
                            IMPORT_google_rpc_status_pb2.Status(),
                        )
                    )
                )
        raise RuntimeError("Unexpected result from Range")

    async def ReverseRange(
        self,
        context: IMPORT_reboot_aio_contexts.ReaderContext,
        state: rbt.std.collections.v1.sorted_map_pb2.SortedMap,
        request: rbt.std.collections.v1.sorted_map_pb2.ReverseRangeRequest,
    ) -> rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse:
        json_state = IMPORT_google_protobuf_json_format.MessageToJson(state)
        json_request = IMPORT_google_protobuf_json_format.MessageToJson(request)

        aborted: IMPORT_asyncio.Future[None] = IMPORT_asyncio.Future()

        try:
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'reader',
                'ReverseRange',
                context,
                aborted,
                json_state,
                json_request,
            )
        except IMPORT_asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = IMPORT_json.loads(json_result)

            if 'response' in result:
                return IMPORT_google_protobuf_json_format.ParseDict(
                    result['response'],
                    rbt.std.collections.v1.sorted_map_pb2.ReverseRangeResponse(),
                )
            elif 'status' in result:
                raise (
                    SortedMap
                    .ReverseRangeAborted
                    .from_status(
                        IMPORT_google_protobuf_json_format.ParseDict(
                            result['status'],
                            IMPORT_google_rpc_status_pb2.Status(),
                        )
                    )
                )
        raise RuntimeError("Unexpected result from ReverseRange")



############################ Reference Node adapters ############################
# Used by Node.js WeakReference implementations to access Python code and
# vice-versa. Relevant to clients.

class SortedMapWeakReferenceNodeAdaptor(SortedMap.WeakReference[SortedMap.WeakReference._Schedule]):

    async def _call(  # type: ignore[override]
        self,
        *,
        callable: IMPORT_typing.Callable[[IMPORT_google_protobuf_message.Message], IMPORT_typing.Awaitable],
        aborted_type: type[IMPORT_reboot.aio.aborted.Aborted],
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        try:
            IMPORT_google_protobuf_json_format.Parse(json_request, request)
            response = await callable(request)
        except IMPORT_google_protobuf_json_format.ParseError as parse_error:
            aborted_error = IMPORT_reboot.aio.aborted.SystemAborted(
                IMPORT_rbt_v1alpha1.errors_pb2.Unknown(),
                message=f"{parse_error}; "
                       "This is usually caused by a deeply nested protobuf message, which is not supported by protobuf.\n"
                        "See the limits here: https://protobuf.dev/programming-guides/proto-limits/"
                )

            return IMPORT_json.dumps(
                {
                    'status': IMPORT_google_protobuf_json_format.MessageToDict(
                        aborted_error.to_status()
                    )
                }
            )
        except BaseException as exception:
            if isinstance(exception, aborted_type):
                return IMPORT_json.dumps(
                    {
                        'status': IMPORT_google_protobuf_json_format.MessageToDict(
                            exception.to_status()
                        )
                    }
                )
            raise
        else:
            return IMPORT_json.dumps(
                {
                    'response': IMPORT_google_protobuf_json_format.MessageToDict(
                        response
                    )
                }
            )

    async def _schedule(  # type: ignore[override]
        self,
        *,
        method: str,
        context: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        schedule: IMPORT_reboot_time_DateTimeWithTimeZone,
        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency],
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        IMPORT_google_protobuf_json_format.Parse(json_request, request)

        if isinstance(context, IMPORT_reboot_aio_contexts.WriterContext):
            task = getattr(
                SortedMapServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ),
                method,
            )(request, schedule=schedule)

            return IMPORT_json.dumps(
                {
                    'taskId': IMPORT_google_protobuf_json_format.MessageToDict(
                        task.task_id
                    )
                }
            )

        # Add scheduling information to the metadata.
        metadata: IMPORT_reboot_aio_types.GrpcMetadata = (
            (IMPORT_reboot_aio_headers.TASK_SCHEDULE, schedule.isoformat()),
        )

        task_id = await getattr(super()._tasks(context), method)(
            request,
            idempotency=idempotency,
            metadata=metadata,
        )

        return IMPORT_json.dumps(
            {
                'taskId': IMPORT_google_protobuf_json_format.MessageToDict(task_id)
            }
        )

    async def _reader(  # type: ignore[override]
        self,
        method: str,
        context: IMPORT_reboot_aio_contexts.ReaderContext | IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = IMPORT_json.loads(json_options)
        method_handle = IMPORT_functools.partial(
            getattr(super()._reader(context), method),
            bearer_token=options.get("bearerToken"),
        )
        return await self._call(
            callable=method_handle,
            aborted_type=getattr(
                SortedMap, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _writer(  # type: ignore[override]
        self,
        method: str,
        context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = IMPORT_json.loads(json_options)

        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = IMPORT_reboot_aio_contexts.Context.idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
                each_iteration=options['idempotency'].get('eachIteration', None),
            )

        if 'schedule' in options:
            when = IMPORT_google_protobuf_timestamp_pb2.Timestamp()
            when.FromJsonString(options['schedule']['when'])
            return await self._schedule(
                method=method,
                context=context,
                schedule=IMPORT_reboot_time_DateTimeWithTimeZone.from_protobuf_timestamp(when),
                idempotency=idempotency,
                request_type=request_type,
                json_request=json_request,
            )

        method_handle = IMPORT_functools.partial(
            getattr(super()._writer(context), method),
            idempotency=idempotency,
            bearer_token=options.get("bearerToken"),
        )
        return await self._call(
            callable=method_handle,
            aborted_type=getattr(
                SortedMap, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _transaction(  # type: ignore[override]
        self,
        method: str,
        context: IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = IMPORT_json.loads(json_options)

        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = IMPORT_reboot_aio_contexts.Context.idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
                each_iteration=options['idempotency'].get('eachIteration', None),
            )

        if 'schedule' in options:
            when = IMPORT_google_protobuf_timestamp_pb2.Timestamp()
            when.FromJsonString(options['schedule']['when'])
            return await self._schedule(
                method=method,
                context=context,
                schedule=IMPORT_reboot_time_DateTimeWithTimeZone.from_protobuf_timestamp(when),
                idempotency=idempotency,
                request_type=request_type,
                json_request=json_request,
            )

        method_handle = IMPORT_functools.partial(
            getattr(super()._workflow(context), method),
            idempotency=idempotency,
            bearer_token=options.get("bearerToken"),
        )
        return await self._call(
            callable=method_handle,
            aborted_type=getattr(
                SortedMap, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _workflow(  # type: ignore[override]
        self,
        method: str,
        context: IMPORT_reboot_aio_contexts.WriterContext | IMPORT_reboot_aio_contexts.TransactionContext | IMPORT_reboot_aio_contexts.WorkflowContext | IMPORT_reboot_aio_external.ExternalContext,
        request_type: type[IMPORT_google_protobuf_message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = IMPORT_json.loads(json_options)

        idempotency: IMPORT_typing.Optional[IMPORT_reboot_aio_idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = IMPORT_reboot_aio_contexts.Context.idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
                each_iteration=options['idempotency'].get('eachIteration', None),
            )

        assert 'schedule' in options

        when = IMPORT_google_protobuf_timestamp_pb2.Timestamp()
        when.FromJsonString(options['schedule']['when'])

        return await self._schedule(
            method=method,
            context=context,
            schedule=IMPORT_reboot_time_DateTimeWithTimeZone.from_protobuf_timestamp(when),
            idempotency=idempotency,
            request_type=request_type,
            json_request=json_request,
        )

# yapf: enable
